from physics.environment.meteorology.base_meteorology import BaseMeteorology
from physics.environment.gis.gis import calculate_path_distances
import numpy as np
from numba import jit
import physics_rs
from typing import Optional
import datetime


class CloudedMeteorology(BaseMeteorology):
    """
    CloudedMeteorology encapsulates meteorological data that includes
    cloud cover, but not solar irradiance (necessitating manual computation).
    """
    def __init__(self, race, weather_forecasts):
        super().__init__()

        self._latitude: Optional[np.ndarray] = None
        self._longitude: Optional[np.ndarray] = None
        self._unix_time: Optional[np.ndarray] = None
        self._cloud_cover: Optional[np.ndarray] = None

        self._race = race
        self._weather_forecast = weather_forecasts

        self.S_0 = 1367.0  # Solar Constant, 1367W/m^2

        self.last_updated_time = self._weather_forecast[0, 0, 2]

    def spatially_localize(self, cumulative_distances: np.ndarray, simplify_weather: bool = False) -> None:
        """

        IMPORTANT: we only have weather coordinates for a discrete set of coordinates. However, the car could be at any
        coordinate in between these available weather coordinates. We need to figure out what coordinate the car is at
        at each timestep and then we can figure out the full weather forecast at each timestep.

        For example, imagine the car is at some coordinate (10, 20). Further imagine that we have a week's worth of
        weather forecasts for the following five coordinates: (5, 4), (11, 19), (20, 30), (40, 30), (0, 60). Which
        set of weather forecasts should we choose? Well, we should choose the (11, 19) one since our coordinate
        (10, 20) is closest to (11, 19). This is what the following code is accomplishing. However, it is not dealing
        with the coordinates directly but rather is dealing with the distances between the coordinates.

        Furthermore, once we have chosen a week's worth of weather forecasts for a specific coordinate, we must isolate
        a single weather forecast depending on what time the car is at the coordinate (10, 20). That is the job of the
        `get_weather_forecast_in_time()` method.

        :param np.ndarray cumulative_distances: NumPy Array representing cumulative distances theoretically achievable for a given input speed array
        :param bool simplify_weather: enable to only use a single weather coordinate (for track races without varying weather)
        """

        # If racing a track race, there is no need for distance calculations. We will return only the origin coordinate
        # This characterizes the weather at every point along the FSGP tracks
        # with the weather at a single coordinate on the track, which is great for reducing the API calls and is a
        # reasonable assumption to make for FSGP only.
        if simplify_weather:
            self._weather_indices = np.zeros_like(cumulative_distances, dtype=int)
            return

        # a list of all the coordinates that we have weather data for
        weather_coords = self._weather_forecast[:, 0, 0:2]

        # distances between all the coordinates that we have weather data for
        weather_path_distances = calculate_path_distances(weather_coords)
        cumulative_weather_path_distances = np.cumsum(weather_path_distances)

        # makes every even-index element negative, this allows the use of np.diff() to calculate the sum of consecutive
        # elements
        cumulative_weather_path_distances[::2] *= -1

        # contains the average distance between two consecutive elements in the cumulative_weather_path_distances array
        average_distances = np.abs(np.diff(cumulative_weather_path_distances) / 2)

        return physics_rs.closest_weather_indices_loop(cumulative_distances, average_distances)

    def temporally_localize(self, unix_timestamps, start_time, tick) -> None:
        """

        Takes in an array of indices of the weather_forecast array, and an array of timestamps. Uses those to figure out
        what the weather forecast is at each time step being simulated.

        we only have weather at discrete timestamps. The car however can be in any timestamp in
        between. Therefore, we must be able to choose the weather timestamp that is closest to the one that the car is in
        so that we can more accurately determine the weather experienced by the car at that timestamp.

        For example, imagine the car is at some coordinate (x,y) at timestamp 100. Imagine we know the weather forecast
        at (x,y) for five different timestamps: 0, 30, 60, 90, and 120. Which weather forecast should we
        choose? Clearly, we should choose the weather forecast at 90 since it is the closest to 100. That's what the
        below code is accomplishing.

        :param np.ndarray unix_timestamps: (int[N]) unix timestamps of the vehicle's journey
        :param int tick: length of a tick in seconds
        :returns:
            - A NumPy array of size [N][9]
            - [9] (latitude, longitude, unix_time, timezone_offset, unix_time_corrected, wind_speed, wind_direction,
                        cloud_cover, precipitation, description):
        :rtype: np.ndarray

        """
        weather_data = physics_rs.weather_in_time(unix_timestamps.astype(np.int64), self._weather_indices.astype(np.int64), self._weather_forecast, 4)
        # roll_by_tick = int(3600 / tick) * (24 + start_hour - hour_from_unix_timestamp(weather_data[0, 2]))
        # weather_data = np.roll(weather_data, -roll_by_tick, 0)

        self._latitude = weather_data[:, 0]
        self._longitude = weather_data[:, 1]
        self._unix_time = weather_data[:, 2]
        self._wind_speed = weather_data[:, 5]
        self._wind_direction = weather_data[:, 6]
        self._cloud_cover = weather_data[:, 7]

    def calculate_solar_irradiances(self, coords, time_zones, local_times, elevations):
        """

        Calculates the Global Horizontal Irradiance from the Sun, relative to a location
        on the Earth, for arrays of coordinates, times, elevations and weathers
        https://www.pveducation.org/pvcdrom/properties-of-sunlight/calculation-of-solar-insolation
        Note: If local_times and time_zones are both unadjusted for Daylight Savings, the
                calculation will end up just the same

        :param np.ndarray coords: (float[N][lat, lng]) array of latitudes and longitudes
        :param np.ndarray time_zones: (int[N]) time zones at different locations in seconds relative to UTC
        :param np.ndarray local_times: (int[N]) unix time that the vehicle will be at each location. (Adjusted for Daylight Savings)
        :param np.ndarray elevations: (float[N]) elevation from sea level in m
        :returns: (float[N]) Global Horizontal Irradiance in W/m2
        :rtype: np.ndarray

        """
        day_of_year, local_time = physics_rs.calculate_array_ghi_times(local_times)

        ghi = self._calculate_GHI(coords[:, 0], coords[:, 1], time_zones,
                                  day_of_year, local_time, elevations, self._cloud_cover)

        stationary_irradiance = self._calculate_angled_irradiance(coords[:, 0], coords[:, 1], time_zones, day_of_year,
                                                                  local_time, elevations, self._cloud_cover)

        # Use stationary irradiance when the car is not driving
        effective_irradiance = np.where(
            np.logical_not(self._race.driving_boolean),
            stationary_irradiance,
            ghi)

        return effective_irradiance

    @staticmethod
    def _calculate_hour_angle(time_zone_utc, day_of_year, local_time, longitude):
        """

        Calculates and returns the Hour Angle of the Sun in the sky.
        https://www.pveducation.org/pvcdrom/properties-of-sunlight/solar-time
        Note: If local time and time_zone_utc are both unadjusted for Daylight Savings, the
                calculation will end up just the same
        :param np.ndarray time_zone_utc: The UTC time zone of your area in hours of UTC offset.
        :param np.ndarray day_of_year: The number of the day of the current year, with January 1 being the first day of the year.
        :param np.ndarray local_time: The local time in hours from midnight. (Adjust for Daylight Savings)
        :param np.ndarray longitude: The longitude of a location on Earth
        :returns: The Hour Angle in degrees.
        :rtype: np.ndarray

        """

        lst = local_time_to_apparent_solar_time(time_zone_utc / 3600, day_of_year,
                                                        local_time, longitude)

        hour_angle = 15 * (lst - 12)

        return hour_angle

    def _calculate_elevation_angle(self, latitude, longitude, time_zone_utc, day_of_year,
                                   local_time):
        """

        Calculates the Elevation Angle of the Sun relative to a location on the Earth
        https://www.pveducation.org/pvcdrom/properties-of-sunlight/elevation-angle
        Note: If local time and time_zone_utc are both unadjusted for Daylight Savings, the
                calculation will end up just the same

        :param np.ndarray latitude: The latitude of a location on Earth
        :param np.ndarray longitude: The longitude of a location on Earth
        :param np.ndarray time_zone_utc: The UTC time zone of your area in hours of UTC offset. For example, Vancouver has time_zone_utc = -7
        :param np.ndarray day_of_year: The number of the day of the current year, with January 1 being the first day of the year.
        :param np.ndarray  local_time: The local time in hours from midnight. (Adjust for Daylight Savings)
        :returns: The elevation angle in degrees
        :rtype: np.ndarray

        """

        # Negative declination angles: Northern Hemisphere winter
        # 0 declination angle : Equinoxes (March 22, Sept 22)
        # Positive declination angle: Northern Hemisphere summer
        declination_angle = calculate_declination_angle(day_of_year)

        # Negative hour angles: Morning
        # 0 hour angle : Solar noon
        # Positive hour angle: Afternoon
        hour_angle = self._calculate_hour_angle(time_zone_utc, day_of_year,
                                                local_time, longitude)
        # From: https://en.wikipedia.org/wiki/Hour_angle#:~:text=At%20solar%20noon%20the%20hour,times%201.5%20hours%20before%20noon).
        # "For example, at 10:30 AM local apparent time
        # the hour angle is −22.5° (15° per hour times 1.5 hours before noon)."

        # mathy part is delegated to a helper function to optimize for numba compilation
        return compute_elevation_angle_math(declination_angle, hour_angle, latitude)

    def _calculate_zenith_angle(self, latitude, longitude, time_zone_utc, day_of_year,
                                local_time):
        """

        Calculates the Zenith Angle of the Sun relative to a location on the Earth
        https://www.pveducation.org/pvcdrom/properties-of-sunlight/azimuth-angle
        Note: If local time and time_zone_utc are both unadjusted for Daylight Savings, the
                calculation will end up just the same

        :param latitude: The latitude of a location on Earth
        :param longitude: The longitude of a location on Earth
        :param time_zone_utc: The UTC time zone of your area in hours of UTC offset.
        :param day_of_year: The number of the day of the current year, with January 1 being the first day of the year.
        :param local_time: The local time in hours from midnight. (Adjust for Daylight Savings)
        :return: The zenith angle in degrees
        :rtype: float

        """

        elevation_angle = self._calculate_elevation_angle(latitude, longitude,
                                                          time_zone_utc, day_of_year, local_time)

        return 90 - elevation_angle

    def _calculate_azimuth_angle(self, latitude, longitude, time_zone_utc, day_of_year,
                                local_time):
        """

        Calculates the Azimuth Angle of the Sun relative to a location on the Earth.
        https://www.pveducation.org/pvcdrom/properties-of-sunlight/azimuth-angle
        Note: If local time and time_zone_utc are both unadjusted for Daylight Savings, the
                calculation will end up just the same

        :param latitude: The latitude of a location on Earth
        :param longitude: The longitude of a location on Earth
        :param time_zone_utc: The UTC time zone of your area in hours of UTC offset. For example, Vancouver has time_zone_utc = -7
        :param day_of_year: The number of the day of the current year, with January 1 being the first day of the year.
        :param local_time: The local time in hours from midnight. (Adjust for Daylight Savings)
        :returns: The azimuth angle in degrees
        :rtype: np.ndarray

        """

        declination_angle = calculate_declination_angle(day_of_year)
        hour_angle = self._calculate_hour_angle(time_zone_utc, day_of_year,
                                                local_time, longitude)

        term_1 = np.sin(np.radians(declination_angle)) * \
            np.sin(np.radians(latitude))

        term_2 = np.cos(np.radians(declination_angle)) * \
            np.sin(np.radians(latitude)) * \
            np.cos(np.radians(hour_angle))

        elevation_angle = self._calculate_elevation_angle(latitude, longitude,
                                                          time_zone_utc, day_of_year, local_time)

        term_3 = np.float_(term_1 - term_2) / \
            np.cos(np.radians(elevation_angle))

        if term_3 < -1:
            term_3 = -1
        elif term_3 > 1:
            term_3 = 1

        azimuth_angle = np.arcsin(term_3)

        return np.degrees(azimuth_angle)

    # ----- Calculation of sunrise and sunset times -----

    # ----- Calculation of modes of solar irradiance -----

    def _calculate_DNI(self, latitude, longitude, time_zone_utc, day_of_year,
                       local_time, elevation):
        """

        Calculates the Direct Normal Irradiance from the Sun, relative to a location
        on the Earth (clearsky)
        https://www.pveducation.org/pvcdrom/properties-of-sunlight/calculation-of-solar-insolation
        Note: If local time and time_zone_utc are both unadjusted for Daylight Savings, the
                calculation will end up just the same

        :param np.ndarray latitude: The latitude of a location on Earth
        :param np.ndarray longitude: The longitude of a location on Earth
        :param np.ndarray time_zone_utc: The UTC time zone of your area in hours of UTC offset.
        :param np.ndarray day_of_year: The number of the day of the current year, with January 1 being the first day of the year.
        :param np.ndarray local_time: The local time in hours from midnight. (Adjust for Daylight Savings)
        :param np.ndarray elevation: The local elevation of a location in metres
        :returns: The Direct Normal Irradiance in W/m2
        :rtype: np.ndarray

        """

        zenith_angle = self._calculate_zenith_angle(latitude, longitude,
                                                    time_zone_utc, day_of_year, local_time)
        a = 0.14

        # https://www.pveducation.org/pvcdrom/properties-of-sunlight/air-mass
        # air_mass = 1 / (math.cos(math.radians(zenith_angle)) + \
        #            0.50572*pow((96.07995 - zenith_angle), -1.6364))

        with np.errstate(invalid="ignore"):
            air_mass = np.float_(1) / (np.float_(np.cos(np.radians(zenith_angle)))
                                       + 0.50572*np.power((96.07995 - zenith_angle), -1.6364))

        with np.errstate(over="ignore"):
            DNI = self.S_0 * ((1 - a * elevation * 0.001) * np.power(0.7, np.power(air_mass, 0.678))
                                  + a * elevation * 0.001)

        return np.where(zenith_angle > 90, 0, DNI)

    def _calculate_DHI(self, latitude, longitude, time_zone_utc, day_of_year,
                       local_time, elevation):
        """

        Calculates the Diffuse Horizontal Irradiance from the Sun, relative to a location
        on the Earth (clearsky)
        https://www.pveducation.org/pvcdrom/properties-of-sunlight/calculation-of-solar-insolation
        Note: If local time and time_zone_utc are both unadjusted for Daylight Savings, the
                calculation will end up just the same

        :param np.ndarray latitude: The latitude of a location on Earth
        :param np.ndarray longitude: The longitude of a location on Earth
        :param np.ndarray time_zone_utc: The UTC time zone of your area in hours of UTC offset.
        :param np.ndarray np.ndarray day_of_year: The number of the day of the current year, with January 1 being the first day of the year.
        :param np.ndarray local_time: The local time in hours from midnight
        :param np.ndarray elevation: The local elevation of a location in metres
        :returns: The Diffuse Horizontal Irradiance in W/m2
        :rtype: np.ndarray

        """

        DNI = self._calculate_DNI(latitude, longitude, time_zone_utc, day_of_year,
                                  local_time, elevation)

        DHI = 0.1 * DNI

        return DHI

    def _calculate_GHI(self, latitude, longitude, time_zone_utc, day_of_year,
                       local_time, elevation, cloud_cover):
        """

        Calculates the Global Horizontal Irradiance from the Sun, relative to a location
        on the Earth
        https://www.pveducation.org/pvcdrom/properties-of-sunlight/calculation-of-solar-insolation
        Note: If local time and time_zone_utc are both unadjusted for Daylight Savings, the
                calculation will end up just the same

        :param np.ndarray latitude: The latitude of a location on Earth
        :param np.ndarray longitude: The longitude of a location on Earth
        :param np.ndarray time_zone_utc: The UTC time zone of your area in hours of UTC offset, without including the effects of Daylight Savings Time. For example, Vancouver has time_zone_utc = -8 year-round.
        :param np.ndarray day_of_year: The number of the day of the current year, with January 1 being the first day of the year.
        :param np.ndarray local_time: The local time in hours from midnight.
        :param np.ndarray elevation: The local elevation of a location in metres
        :param np.ndarray cloud_cover: A NumPy array representing cloud cover as a percentage from 0 to 100
        :returns: The Global Horizontal Irradiance in W/m^2
        :rtype: np.ndarray

        """

        DHI = self._calculate_DHI(latitude, longitude, time_zone_utc, day_of_year,
                                  local_time, elevation)

        DNI = self._calculate_DNI(latitude, longitude, time_zone_utc, day_of_year,
                                  local_time, elevation)

        zenith_angle = self._calculate_zenith_angle(latitude, longitude,
                                                    time_zone_utc, day_of_year, local_time)

        GHI = DNI * np.cos(np.radians(zenith_angle)) + DHI

        return self._apply_cloud_cover(GHI=GHI, cloud_cover=cloud_cover)

    @staticmethod
    def _apply_cloud_cover(GHI, cloud_cover):
        """

        Applies a cloud cover model to the GHI data.

        Cloud cover adjustment follows the equation laid out here:
        http://www.shodor.org/os411/courses/_master/tools/calculators/solarrad/

        :param np.ndarray GHI: Global Horizontal Index in W/m^2
        :param np.ndarray cloud_cover: A NumPy array representing cloud cover as a percentage from 0 to 100

        :returns: GHI after considering cloud cover data
        :rtype: np.ndarray

        """

        assert np.logical_and(cloud_cover >= 0, cloud_cover <= 100).all()

        scaled_cloud_cover = cloud_cover / 100

        assert np.logical_and(scaled_cloud_cover >= 0,
                              scaled_cloud_cover <= 1).all()

        return GHI * (1 - (0.75 * np.power(scaled_cloud_cover, 3.4)))

    # ----- Calculation of modes of solar irradiance, but returning numpy arrays -----
    @staticmethod
    def _date_convert(date):
        """

        Convert a date into local time.

        :param datetime.datetime date: date to be converted
        :return: a date converted into local time.
        :rtype: int

        """

        return date.hour + (float(date.minute * 60 + date.second) / 3600)

    def _calculate_angled_irradiance(self, latitude, longitude, time_zone_utc, day_of_year,
                                     local_time, elevation, cloud_cover, array_angles=np.array([0, 15, 30, 45])):
        """

        Determine the direct and diffuse irradiance on an array which can be mounted at different angles.
        During stationary charging, the car can mount the array at different angles, resulting in a higher
        component of direct irradiance captured.

        Uses the GHI formula, GHI = DNI*cos(zenith)+DHI but with an 'effective zenith',
        the angle between the mounted panel's normal and the sun.

        :param np.ndarray latitude: The latitude of a location on Earth
        :param np.ndarray longitude: The longitude of a location on Earth
        :param np.ndarray time_zone_utc: The UTC time zone of your area in hours of UTC offset, without including the effects of Daylight Savings Time. For example, Vancouver has time_zone_utc = -8 year-round.
        :param np.ndarray day_of_year: The number of the day of the current year, with January 1 being the first day of the year.
        :param np.ndarray local_time: The local time in hours from midnight.
        :param np.ndarray elevation: The local elevation of a location in metres
        :param np.ndarray cloud_cover: A NumPy array representing cloud cover as a percentage from 0 to 100
        :param np.ndarray array_angles: An array containing the discrete angles on which the array can be mounted
        :returns: The "effective Global Horizontal Irradiance" in W/m^2
        :rtype: np.ndarray

        """

        DHI = self._calculate_DHI(latitude, longitude, time_zone_utc, day_of_year,
                                  local_time, elevation)

        DNI = self._calculate_DNI(latitude, longitude, time_zone_utc, day_of_year,
                                  local_time, elevation)

        zenith_angle = self._calculate_zenith_angle(latitude, longitude,
                                                    time_zone_utc, day_of_year, local_time)

        # Calculate the absolute differences
        differences = np.abs(zenith_angle[:, np.newaxis] - array_angles)

        # Find the minimum difference for each element in zenith_angle
        effective_zenith = np.min(differences, axis=1)

        # Now effective_zenith contains the minimum absolute difference for each element in zenith_angle

        GHI = DNI * np.cos(np.radians(effective_zenith)) + DHI

        return self._apply_cloud_cover(GHI=GHI, cloud_cover=cloud_cover)


def local_time_to_apparent_solar_time(time_zone_utc, day_of_year, local_time,
                                      longitude):
    """

    Converts between the local time to the apparent solar time and returns the apparent
    solar time.
    https://www.pveducation.org/pvcdrom/properties-of-sunlight/solar-time

    Note: If local time and time_zone_utc are both unadjusted for Daylight Savings, the
        calculation will end up just the same

    :param np.ndarray time_zone_utc: The UTC time zone of your area in hours of UTC offset.
    :param np.ndarray day_of_year: The number of the day of the current year, with January 1 being the first day of the year.
    :param np.ndarray local_time: The local time in hours from midnight (Adjust for Daylight Savings)
    :param np.ndarray longitude: The longitude of a location on Earth
    :returns: The Apparent Solar Time of a location, in hours from midnight
    :rtype: np.ndarray

    """

    lstm = calculate_LSTM(time_zone_utc)
    eot = calculate_eot_correction(day_of_year)

    # local solar time
    lst = local_time + np.float_(longitude - lstm) / 15 + np.float_(eot) / 60

    return lst


@jit(nopython=True)
def calculate_LSTM(time_zone_utc):
    """

    Calculates and returns the LSTM, or Local Solar Time Meridian.
    https://www.pveducation.org/pvcdrom/properties-of-sunlight/solar-time

    :param np.ndarray time_zone_utc: The UTC time zone of your area in hours of UTC offset.
    :returns: The Local Solar Time Meridian in degrees
    :rtype: np.ndarray

    """

    return 15 * time_zone_utc


@jit(nopython=True)
def calculate_eot_correction(day_of_year):
    """

    Approximates and returns the correction factor between the apparent
    solar time and the mean solar time

    :param np.ndarray day_of_year: The number of the day of the current year, with January 1 being the first day of the year.
    :returns: The Equation of Time correction EoT in minutes, where apparent Solar Time = Mean Solar Time + EoT
    :rtype: np.ndarray

    """

    b = np.radians((np.float_(360) / 364) * (day_of_year - 81))

    eot = 9.87 * np.sin(2 * b) - 7.83 * np.cos(b) - 1.5 * np.sin(b)

    return eot


def get_day_of_year_map(date):
    """

    Extracts day, month, year, from datetime object

    :param datetime.date date: date to be decomposed

    """
    return get_day_of_year(date.day, date.month, date.year)


def get_day_of_year(day, month, year):
    """

    Calculates the day of the year, given the day, month and year.
    Day refers to a number representing the nth day of the year. So, Jan 1st will be the 1st day of the year

    :param int day: nth day of the year
    :param int month: month
    :param int year: year
    :returns: day of year
    :rtype: int

    """

    return (datetime.date(year, month, day) - datetime.date(year, 1, 1)).days + 1


@jit(nopython=True)
def calculate_declination_angle(day_of_year):
    """

    Calculates the Declination Angle of the Earth at a given day
    https://www.pveducation.org/pvcdrom/properties-of-sunlight/declination-angle

    :param np.ndarray day_of_year: The number of the day of the current year, with January 1 being the first day of the year.
    :returns: The declination angle of the Earth relative to the Sun, in degrees
    :rtype: np.ndarray

    """

    declination_angle = -23.45 * np.cos(np.radians((np.float_(360) / 365) *
                                                   (day_of_year + 10)))

    return declination_angle


@jit(nopython=True)
def compute_elevation_angle_math(declination_angle, hour_angle, latitude):
    """

    Gets the two terms to calculate and return elevation angle, given the
    declination angle, hour angle, and latitude.

    This method separates the math part of the calculation from its caller
    method to optimize for numba compilation.

    :param np.ndarray latitude: array of latitudes
    :param np.ndarray declination_angle: The declination angle of the Earth relative to the Sun
    :param np.ndarray hour_angle: The hour angle of the sun in the sky
    :returns: The elevation angle in degrees
    :rtype: np.ndarray

    """

    term_1 = np.sin(np.radians(declination_angle)) * np.sin(np.radians(latitude))
    term_2 = np.cos(np.radians(declination_angle)) * np.cos(np.radians(latitude)) * np.cos(np.radians(hour_angle))
    elevation_angle = np.arcsin(term_1 + term_2)

    return np.degrees(elevation_angle)




import physics_rs
import numpy as np
import pytest


def test_calculate_driving_speeds_simple():
    """
    Test a simple case of the driving speeds with no rounding
    """
    average_speeds = np.array([10, 14, 30, 0, 10, 20, 30], dtype=np.float64)
    track_length = 30
    simulation_dt = 1

    idle_time = 3
    driving_allowed = np.array([0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=bool)
    driving_speeds = physics_rs.get_driving_speeds(average_speeds, simulation_dt, driving_allowed, track_length, idle_time)

    assert np.all(np.equal(driving_speeds, [0, 10, 10, 10, 14, 14, 0, 0, 0, 0, 0, 30, 0, 0, 0, 10, 10, 10, 20, 20, 30]))


def test_calculate_driving_speeds_fp():
    """
    Test the driving speeds with floating point rounding going on
    :return:
    """
    average_speeds = np.array([9.9, 30.7, 15.2, 15.3], dtype=np.float64)
    track_length = 30.6
    simulation_dt = 1

    idle_time = 3
    driving_allowed = np.array([0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=bool)
    driving_speeds = physics_rs.get_driving_speeds(average_speeds, simulation_dt, driving_allowed, track_length, idle_time)

    assert np.allclose(driving_speeds, np.array([0, 0, 9.9, 9.9, 9.9, 9.9, 30.7, 15.2, 15.2, 15.2, 15.3, 15.3]), atol=1e-3)


def test_calculate_driving_speeds_fp_with_tick():
    """
    Test the driving speeds with floating point rounding going on and tick != 1.
    """
    average_speeds = np.array([10.2, 30.7, 15.2, 15.3], dtype=np.float64)
    track_length = 30.6
    simulation_dt = 2

    idle_time = 3
    driving_allowed = np.array([0, 0, 1, 1, 1, 1, 1, 1], dtype=bool)
    driving_speeds = physics_rs.get_driving_speeds(average_speeds, simulation_dt, driving_allowed, track_length, idle_time)

    assert np.allclose(driving_speeds, np.array([0, 0, 10.2, 10.2, 30.7, 15.2, 15.2, 15.3]), atol=1e-3)
    pass


def test_failure():
    """
    Test that nothing fails when too many average speeds are provided, and that ValueError is raised when
    not enough laps are provided.
    """
    average_speeds_too_many = np.array([10, 14, 30, 0, 10, 20, 30, 10, 10], dtype=np.float64)
    average_speeds_too_few = np.array([10, 14, 30, 0, 10, 20], dtype=np.float64)
    track_length = 30
    simulation_dt = 1
    idle_time = 3
    driving_allowed = np.array([0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=bool)

    with pytest.raises(ValueError):
        physics_rs.get_driving_speeds(average_speeds_too_few, simulation_dt, driving_allowed, track_length, idle_time)

    driving_speeds = physics_rs.get_driving_speeds(average_speeds_too_many, simulation_dt, driving_allowed, track_length, idle_time)

    assert np.all(np.equal(driving_speeds, [0, 10, 10, 10, 14, 14, 0, 0, 0, 0, 0, 30, 0, 0, 0, 10, 10, 10, 20, 20, 30]))



import pytest
import numpy as np
from unittest.mock import Mock
from physics.models.battery import FilteredBatteryModel, KalmanFilterConfig


@pytest.fixture
def mock_config():
    # Mock the battery model config
    battery_model_config = Mock()
    battery_model_config.Q_total = 3600.0

    battery_model_config.get_Uoc = lambda soc: 3.5 + 0.5 * soc
    battery_model_config.get_R_0 = lambda soc: 0.01 + 0.005 * soc
    battery_model_config.get_R_P = lambda soc: 0.02 + 0.002 * soc
    battery_model_config.get_C_P = lambda soc: 1000.0 + 100.0 * soc

    # Mock the full filtered battery model config
    config = Mock()
    config.battery_model_config = battery_model_config
    config.state_covariance_matrix = np.eye(2) * 0.01
    config.process_noise_matrix = np.eye(2) * 1e-6
    config.measurement_noise_vector = np.array([[0.001]])

    return config


def test_initialization(mock_config):
    model = FilteredBatteryModel(mock_config, initial_SOC=0.9, initial_Uc=0.1)
    assert np.isclose(model.SOC, 0.9)
    assert np.isclose(model.Uc, 0.1)
    assert model.Ut == 0


def test_predict_then_update_changes_state(mock_config):
    model = FilteredBatteryModel(mock_config)
    SOC_before = model.SOC
    Uc_before = model.Uc

    # Run a prediction and update step
    model.predict_then_update(measured_Ut=3.7, current=2.0, time_step=1.0)

    SOC_after = model.SOC
    Uc_after = model.Uc

    # Ensure state is updated
    assert not np.isclose(SOC_before, SOC_after)
    assert not np.isclose(Uc_before, Uc_after)


def test_failure(mock_config):
    with pytest.raises(AssertionError):
        model = FilteredBatteryModel(mock_config, initial_SOC=1.5)

    with pytest.raises(AssertionError):
        model = FilteredBatteryModel(mock_config, initial_SOC=-0.2)

    with pytest.raises(AssertionError):
        model = FilteredBatteryModel(mock_config, initial_SOC=1.0, alpha=-0.1)

    with pytest.raises(AssertionError):
        model = FilteredBatteryModel(mock_config, initial_SOC=1.0, alpha=1.1)


def test_measurement_function_and_jacobian_shape(mock_config):
    model = FilteredBatteryModel(mock_config)
    x = np.array([0.8, 0.05])
    H = model._measurement_jacobian(x)
    z = model._measurement_function(x)

    assert H.shape == (1, 2)
    assert isinstance(z, float)



from .gis import (
    GIS,
)

from .meteorology import (
    IrradiantMeteorology,
    CloudedMeteorology,
    BaseMeteorology
)

__all__ = [
    "IrradiantMeteorology",
    "CloudedMeteorology",
    "GIS",
]



from .arrays import BasicArray
from .battery import BasicBattery
from .lvs import BasicLVS
from .motor import BasicMotor
from .regen import BasicRegen

__all__ = [
    "BasicArray",
    "BasicBattery",
    "BasicLVS",
    "BasicMotor",
    "BasicRegen"
]



import logging
import math
import physics_rs
import numpy as np
import sys

from numpy.typing import ArrayLike, NDArray
from tqdm import tqdm
from xml.dom import minidom
from haversine import haversine, Unit
from physics.environment.gis.base_gis import BaseGIS


class GIS(BaseGIS):
    def __init__(self, route_data, origin_coord, current_coord=None):
        """

        Initialises a GIS (geographic location system) object. This object is responsible for getting the
        simulation's planned route from the Google Maps API and performing operations on the received data.

        Requires a map, ``route_data`` with certain keys.
            1. "path": an iterable of shape [N, 2] representing N coordinates in the form (latitude, longitude).
            2. "elevations": an iterable of shape [N] where each Nth element is the elevation, in meters, of the Nth path coordinate.
            3. "time_zones": an iterable of shape [N] where each Nth element is the UTC time zone offset of the Nth path coordinate.
            4. "num_unique_coords": the number of unique coordinates (that is, if the path is a single lap that has been tiled, how many path coordinates compose a single lap).

        :param route_data: map of data containing "path", "elevations", "time_zones", and "num_unique_coords".
        :param origin_coord: NumPy array containing the start coordinate (lat, long) of the planned travel route

        """
        self.path = route_data['path']
        self.launch_point = route_data['path'][0]
        self.path_elevations = route_data['elevations']
        self.path_time_zones = route_data['time_zones']
        self.num_unique_coords = route_data['num_unique_coords']

        if current_coord is not None:
            if not np.array_equal(current_coord, origin_coord):
                logging.warning("Current position is not origin position. Modifying path data.\n")

                # We need to find the closest coordinate along the path to the vehicle position
                current_coord_index = GIS._find_closest_coordinate_index(current_coord, self.path)

                # All coords before the current coordinate should be discarded
                self.path = self.path[current_coord_index:]
                self.path_elevations = self.path_elevations[current_coord_index:]
                self.path_time_zones = self.path_time_zones[current_coord_index:]

        self.path_distances = calculate_path_distances(self.path)
        self.path_length = np.cumsum(calculate_path_distances(self.path[:self.num_unique_coords]))[-1]
        self.path_gradients = calculate_path_gradients(self.path_elevations, self.path_distances)

    @staticmethod
    def process_KML_file(route_file):
        """

        Load the FSGP Track from a KML file exported from a Google Earth project.

        Ensure to follow guidelines enumerated in this directory's `README.md` when creating and
        loading new route files.

        :return: Array of N coordinates (latitude, longitude) in the shape [N][2].
        """
        with open(route_file) as f:
            data = minidom.parse(f)
            kml_coordinates = data.getElementsByTagName("coordinates")[0].childNodes[0].data
            coordinates: np.ndarray = np.array(parse_coordinates_from_kml(kml_coordinates))

            # Google Earth exports coordinates in order longitude, latitude, when we want the opposite
            return np.roll(coordinates, 1, axis=1)

    def calculate_closest_gis_indices(self, distances):
        """

        Takes in an array of point distances from starting point, returns a list of
        ``self.path`` indices of coordinates which have a distance from the starting point
        closest to the point distances.

        :param np.ndarray distances: (float[N]) array of distances, where cumulative_distances[x] > cumulative_distances[x-1]
        :returns: (float[N]) array of indices of path
        :rtype: np.ndarray

        """
        return physics_rs.closest_gis_indices_loop(distances, self.path_distances)

    def calculate_driving_speeds(
            self,
            average_lap_speeds: ArrayLike,
            simulation_dt: int,
            driving_allowed: ArrayLike,
            idle_time: int
    ) -> NDArray[float]:
        """
        Generate valid driving speeds as a simulation-time array given a set of average speeds for each
        simulated lap.
        Driving speeds will only be non-zero when we are allowed to drive, and the speed
        for every tick during a lap will be that lap's corresponding desired average speed for as long
        as it takes to complete the lap.

        :param average_lap_speeds: An array of average speeds in m/s, one for each simulated lap.
            If there are more speeds given than laps available, the unused speeds will be silently ignored.
            If there are too few, an error will be returned.
        :param simulation_dt: The simulated tick length.
        :param driving_allowed: A simulation-time boolean where the `True` elements are when we
            are allowed to drive, and `False` is when we are not. Requires that (at least) the first element is
            `False` due to the race beginning in the morning before we are allowed to drive.
        :param idle_time: The length of time to pause driving upon processing a "0m/s" average speed.
        :return: A simulation-time array of driving speeds in m/s, or an error if there weren't enough
            laps provided to fill the entire simulation time.
        """
        return physics_rs.get_driving_speeds(
            np.array(average_lap_speeds).astype(np.float64),
            simulation_dt,
            np.array(driving_allowed).astype(bool),
            self.path_length,
            idle_time
        )

    @staticmethod
    def _python_calculate_closest_gis_indices(distances, path_distances):
        """

        Python implementation of use_compiled core.closest_gis_indices_loop. See parent function for documentation details.

        """

        current_coordinate_index = 0
        result = []

        with tqdm(total=len(distances), file=sys.stdout, desc="Calculating closest GIS indices") as pbar:
            distance_travelled = 0
            for distance in np.nditer(distances):
                distance_travelled += distance

                while distance_travelled > path_distances[current_coordinate_index]:
                    distance_travelled -= path_distances[current_coordinate_index]
                    current_coordinate_index += 1

                    if current_coordinate_index >= len(path_distances) - 1:
                        current_coordinate_index = len(path_distances) - 1

                result.append(current_coordinate_index)
                pbar.update(1)

        return np.array(result)

    # ----- Getters -----
    def get_time_zones(self, gis_indices):
        """

        Takes in an array of path indices, returns the time zone at each index

        :param np.ndarray gis_indices: (float[N]) array of path indices
        :returns: (float[N]) array of time zones in seconds
        :rtype: np.ndarray

        """

        return self.path_time_zones[gis_indices]

    def get_gradients(self, gis_indices):
        """

        Takes in an array of path indices, returns the road gradient at each index

        :param np.ndarray gis_indices: (float[N]) array of path indices
        :returns: (float[N]) array of road gradients
        :rtype np.ndarray:

        """

        return self.path_gradients[gis_indices]

    def get_path(self):
        """
        Returns all N coordinates of the path in a NumPy array
        [N][latitude, longitude]

        :rtype: np.ndarray

        """

        return self.path

    def get_path_elevations(self):
        """

        Returns all N elevations of the path in a NumPy array
        [N][elevation]

        :rtype: np.ndarray

        """

        return self.path_elevations

    def get_path_distances(self):
        """

        Returns all N-1 distances of the path in a NumPy array
        [N-1][elevation]

        :rtype: np.ndarray

        """

        return self.path_distances

    def get_path_gradients(self):
        """

        Returns all N-1 gradients of a path in a NumPy array
        [N-1][gradient]

        :rtype: np.ndarray

        """

        return self.path_gradients

    # ----- Path calculation functions -----
    def calculate_path_min_max(self):
        logging.warning(f"Using deprecated function 'calculate_path_min_max()'!")
        min_lat, min_long = self.path.min(axis=0)
        max_lat, max_long = self.path.max(axis=0)
        return [min_long, min_lat, max_long, max_lat]

    def calculate_current_heading_array(self):
        """

        Calculates the bearing of the vehicle between consecutive points
        https://www.movable-type.co.uk/scripts/latlong.html

        :returns: array of bearings
        :rtype: np.ndarray

        """
        bearing_array = np.zeros(len(self.path))

        for index in range(0, len(self.path) - 1):
            coord_1 = np.radians(self.path[index])
            coord_2 = np.radians(self.path[index + 1])

            y = math.sin(coord_2[1] - coord_1[1]) \
                * math.cos(coord_2[0])

            x = math.cos(coord_1[0]) \
                * math.sin(coord_2[0]) \
                - math.sin(coord_1[0]) \
                * math.cos(coord_2[0]) \
                * math.cos(coord_2[1] - coord_1[1])

            theta = math.atan2(y, x)

            bearing_array[index] = ((theta * 180) / math.pi + 360) % 360

        bearing_array[-1] = bearing_array[-2]

        return bearing_array

    @staticmethod
    def _calculate_vector_square_magnitude(vector):
        """

        Calculate the square magnitude of an input vector. Must be one-dimensional.

        :param np.ndarray vector: NumPy array[N] representing a vector[N]
        :return: square magnitude of the input vector
        :rtype: float

        """

        return sum(i ** 2 for i in vector)

    @staticmethod
    def _find_closest_coordinate_index(current_coord, path):
        """

        Returns the closest coordinate to current_coord in path

        :param np.ndarray current_coord: A NumPy array[N] representing a N-dimensional vector
        :param np.ndarray path: A NumPy array[M][N] of M coordinates which should be N-dimensional vectors
        :returns: index of the closest coordinate.
        :rtype: int

        """

        to_current_coord_from_path = np.abs(path - current_coord)
        distances_from_current_coord = np.zeros(len(to_current_coord_from_path))
        for i in range(len(to_current_coord_from_path)):
            # As we just need the minimum, using square magnitude will save performance
            distances_from_current_coord[i] = GIS._calculate_vector_square_magnitude(to_current_coord_from_path[i])

        return distances_from_current_coord.argmin()


def calculate_path_distances(coords):
    """

    Obtain the distance between each coordinate by approximating the spline between them
    as a straight line, and use the Haversine formula (https://en.wikipedia.org/wiki/Haversine_formula)
    to calculate distance between coordinates on a sphere.

    :param np.ndarray coords: A NumPy array [n][latitude, longitude]
    :returns path_distances: a NumPy array [n-1][distances],
    :rtype: np.ndarray

    """

    coords_offset = np.roll(coords, (1, 1))
    path_distances = []
    for u, v in zip(coords, coords_offset):
        path_distances.append(haversine(u, v, unit=Unit.METERS))

    return np.array(path_distances)


def parse_coordinates_from_kml(coords_str: str) -> np.ndarray:
    """

    Parse a coordinates string from a XML (KML) file into a list of coordinates (2D vectors).
    Requires coordinates in the format "39.,41.,0  39.,40.,0" which will return [ [39., 41.], [39., 40.] ].

    :param coords_str: coordinates string from a XML (KML) file
    :return: list of 2D vectors representing coordinates
    :rtype: np.ndarray

    """

    def parse_coord(pair):
        coord = pair.split(',')
        coord.pop()
        coord = [float(value) for value in coord]
        return coord

    return list(map(parse_coord, coords_str.split()))


def calculate_path_gradients(elevations, distances):
    """

    Get the approximate gradients of every point on the path.

    Note:
        - gradient > 0 corresponds to uphill
        - gradient < 0 corresponds to downhill

    :param np.ndarray elevations: [N][elevations]
    :param np.ndarray distances: [N-1][distances]
    :returns gradients: [N-1][gradients]
    :rtype: np.ndarray

    """

    # subtract every next elevation with the previous elevation to
    # get the difference in elevation
    # [1 2 3 4 5]
    # [5 1 2 3 4] -
    # -------------
    #   [1 1 1 1]

    offset = np.roll(elevations, 1)
    delta_elevations = elevations - offset

    # Divide the difference in elevation to get the gradient
    # gradient > 0: uphill
    # gradient < 0: downhill
    with np.errstate(invalid='ignore'):
        gradients = delta_elevations / distances

    return np.nan_to_num(gradients, nan=0.)



from physics.models.lvs.base_lvs import BaseLVS


class BasicLVS(BaseLVS):

    def __init__(self, consumed_energy, lvs_current, lvs_voltage):
        super().__init__(consumed_energy)
        self.lvs_current = lvs_current
        self.lvs_voltage = lvs_voltage

    def get_consumed_energy(self, tick):
        """
            Get the energy consumption of the Low Voltage System (current * voltage * time)

            :param tick - (int) tick time passed
            :returns: consumed_energy - (number) value of energy consumed
        """
        return self.lvs_current * self.lvs_voltage * tick



# These tests are required to satisfy STG Specification #2: Physics Publishing

import importlib


def test_version():
    version_module = importlib.import_module("physics")
    assert isinstance(version_module.__version__, str)



import tomli as tomllib
import pathlib
from scipy import optimize
import numpy as np
from physics.models.battery import SOCDependent
from typing import cast
from numpy.typing import NDArray


class BatteryModelConfig:
    """
    A concrete implementation of the `EquivalentCircuitModelConfig` protocol.

    This implementation fits values of U_oc, R_0, R_P, and C_P at various state-of-charge (SOC) values
    to a seventh degree polynomial to generate a smooth function mapping SOC to each battery parameter.

    For example, R_0 = R_0_data[i] when Soc = Soc_data[i].
    """
    def __init__(self, R_0_data, Soc_data, R_P_data, C_P_data, Uoc_data, Q_total):
        # ----- Initialize Parameters -----
        def quintic_polynomial(x, x0, x1, x2, x3, x4, x5, x6, x7):
            return np.polyval(np.array([x0, x1, x2, x3, x4, x5, x6, x7]), x)

        self._U_oc_coefficients, _ = optimize.curve_fit(quintic_polynomial, Soc_data, Uoc_data)
        self._R_0_coefficients, _ = optimize.curve_fit(quintic_polynomial, Soc_data, R_0_data)
        self._C_P_coefficients, _ = optimize.curve_fit(quintic_polynomial, Soc_data, C_P_data)
        self._R_P_coefficients, _ = optimize.curve_fit(quintic_polynomial, Soc_data, R_P_data)

        # Casts are just for the type-checker to know that np.polyval will work as SOCDependent
        self._U_oc: SOCDependent = cast(SOCDependent, lambda soc: np.polyval(self._U_oc_coefficients, soc))  # V
        self._R_0: SOCDependent = cast(SOCDependent, lambda soc: np.polyval(self._R_0_coefficients, soc))    # Ohms
        self._R_P: SOCDependent = cast(SOCDependent, lambda soc: np.polyval(self._R_P_coefficients, soc))    # Ohms
        self._C_P: SOCDependent = cast(SOCDependent, lambda soc: np.polyval(self._C_P_coefficients, soc))    # Farads

        self._Q_total = Q_total

    @property
    def get_Uoc(self) -> SOCDependent:
        return self._U_oc

    @property
    def get_R_0(self) -> SOCDependent:
        return self._R_0

    @property
    def get_R_P(self) -> SOCDependent:
        return self._R_P

    @property
    def get_C_P(self) -> SOCDependent:
        return self._C_P

    @property
    def Q_total(self) -> float:
        return self._Q_total


class KalmanFilterConfig:
    def __init__(
            self,
            battery_model_config: BatteryModelConfig,
            process_noise_matrix: NDArray,
            state_covariance_matrix: NDArray,
            measurement_noise_vector: NDArray
    ):
        self._battery_model_config = battery_model_config
        self._process_noise_matrix = process_noise_matrix
        self._state_covariance_matrix = state_covariance_matrix
        self._measurement_noise_vector = measurement_noise_vector

    @property
    def battery_model_config(self) -> BatteryModelConfig:
        """
        Configuration of the underlying `EquivalentCircuitModel`.
        """
        return self._battery_model_config

    @property
    def process_noise_matrix(self) -> NDArray[float]:
        """
        A 2x2 matrix containing the process noise covariance matrix where [0, 0] is the SOC evolution
        noise and [1, 1] is the polarization potential evolution noise.
        """
        return self._process_noise_matrix

    @property
    def state_covariance_matrix(self) -> NDArray[float]:
        """
        A 2x2 matrix containing the state covariance matrix where [0, 0] is the SOC covariance
        noise and [1, 1] is the polarization potential covariance.
        """
        return self._state_covariance_matrix

    @property
    def measurement_noise_vector(self) -> NDArray[float]:
        """
        A 1x1 vector containing the noise expected in the terminal voltage measurement.
        """
        return self._measurement_noise_vector


def load_battery_config(absolute_path: str | pathlib.Path) -> BatteryModelConfig:
    # Build the full path to the config file
    full_path = pathlib.Path(absolute_path)
    with open(full_path, 'rb') as f:
        data = tomllib.load(f)
    return BatteryModelConfig(**data)



from physics.models.arrays.base_array import BaseArray


class BasicArray(BaseArray):

    # incident_sunlight:
    def __init__(self, panel_efficiency, panel_size):
        super().__init__()

        # solar cell efficiency
        self.panel_efficiency = panel_efficiency

        # solar panel size in m^2
        self.panel_size = panel_size

        # please do not use this.
        self.solar_irradiance = 1200

    def calculate_produced_energy(self, solar_irradiance, tick):
        """

        Returns a numpy array with the energy produced by the solar panels across
        each the length of each tick

        :param np.ndarray solar_irradiance: (float[N]) the global horizontal irradiance(GHI) at
            each moment experienced by the vehicle, in W/m2
        :param float tick: (float) the duration of a time step in seconds

        :returns: (float[N]) array of energy produced by the solar panels on the vehicle
            in Joules
        :rtype: np.ndarray

        """
        return solar_irradiance * self.panel_efficiency * self.panel_size * tick

    def __str__(self):
        return(f"BasicArray: incident_sunlight: {self.solar_irradiance}W/m^2\n"
               f"BasicArray: panel_size: {self.panel_size}m^2\n"
               f"BasicArray: panel_efficiency: {self.panel_efficiency * 100}%\n")



import numpy as np
import pandas as pd
import pathlib
import matplotlib.pyplot as plt
from physics.models.battery import EquivalentCircuitBatteryModel, BatteryModelConfig, load_battery_config


# This test requires a voltage.csv and current.csv in the same directory to run
def csv_to_timeseries_tuples(csv_file):
    path = pathlib.Path(__file__).parent / csv_file
    df = pd.read_csv(path)
    df['Time'] = pd.to_datetime(df['Time'])
    return np.array(list(zip(df['Time'].dt.to_pydatetime(), df['Value'])))


def plot_results(soc_array, predicted_ut_array, voltage_data, window=None):
    fig, ax = plt.subplots()

    if window is None:
        window = slice(0, len(predicted_ut_array), 1)

    ax.plot(predicted_ut_array, label=r"Predicted $U_t$", color="tab:red")
    ax.plot(voltage_data, label=r"Measured $U_t$", color="tab:orange")
    ax.set_ylim(75, 140)
    ax.set_xticks([])
    ax.set_ylabel("Voltage")

    ax2 = ax.twinx()
    ax2.plot(soc_array[window], color="tab:blue", label="Filtered SOC")
    ax2.set_ylabel("SOC")

    ax.legend(loc='upper right')
    ax2.legend(loc='lower left')

    plt.title("Simulation of first-order Thevenin equivalent battery model")
    plt.show()


def battery_model():
    voltage_data = csv_to_timeseries_tuples('voltage.csv')
    current_data = csv_to_timeseries_tuples('current.csv')

    # This dataset has 0.1s period between measurements
    time_difference = 0.1

    current_raw = current_data[:, 1]
    current_error = np.polyval([-0.00388, 1547], current_raw * 1000.0)

    current = current_raw - (current_error / 1000)
    voltage = voltage_data[:, 1]

    energy_array = current * voltage * time_difference

    model_config: BatteryModelConfig = load_battery_config(pathlib.Path(__file__).parent / 'battery_config.toml')

    battery_model = EquivalentCircuitBatteryModel(model_config, state_of_charge=1.04)

    soc_array, predicted_ut_array = battery_model.update_array(tick=time_difference, current_array=np.array(-current, dtype=float))
    # soc_array, predicted_ut_array = battery_model.update_array(tick=time_difference, delta_energy_array=np.array(-energy_array, dtype=float))

    # example usage
    plot_results(soc_array, predicted_ut_array, voltage)


if __name__ == '__main__':
    battery_model()



import numpy as np


def constrain_speeds(speed_limits: np.ndarray, speeds: np.ndarray, tick: int) -> np.ndarray:
    """
    Constrains the vehicle speeds based on the speed limits and computes new speeds.

    :param speed_limits: Array of speed limits (km/h) for each point.
    :param speeds: Array of vehicle speeds (km/h).
    :param tick: The time step (in some unit, e.g., seconds or ticks).

    :return: A NumPy array of constrained vehicle speeds (km/h).
    """
    ...


def calculate_array_ghi_times(python_local_times: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Calculates the array of GHI times based on local times.

    :param python_local_times: Array of local times (UNIX timestamps or other format).

    :return: A tuple of two NumPy arrays:
        - Day of year (f64)
        - Local time (f64)
    """
    ...


def closest_gis_indices_loop(python_cumulative_distances: np.ndarray,
                             python_average_distances: np.ndarray) -> np.ndarray:
    """
    Finds the closest GIS indices based on cumulative and average distances.

    :param python_cumulative_distances: Array of cumulative distances.
    :param python_average_distances: Array of average distances.

    :return: A NumPy array of indices (i64) corresponding to the closest GIS indices.
    """
    ...


def closest_weather_indices_loop(
        python_cumulative_distances: np.ndarray,
        python_average_distances: np.ndarray
) -> np.ndarray:
    """
    Finds the closest weather indices based on cumulative and average distances.

    :param python_cumulative_distances: Array of cumulative distances.
    :param python_average_distances: Array of average distances.

    :return: A NumPy array of indices (i64) corresponding to the closest weather indices.
    """
    ...


def weather_in_time(
        python_unix_timestamps: np.ndarray,
        python_indices: np.ndarray,
        python_weather_forecast: np.ndarray,
        index: int
) -> np.ndarray:
    """
    Retrieves the weather forecast at specific times for given indices.

    :param python_unix_timestamps: Array of UNIX timestamps.
    :param python_indices: Array of indices to look up.
    :param python_weather_forecast: Array of weather forecasts.
    :param index: A specific index to look up in the weather forecast.

    :return: A NumPy array of weather values (f64) at the specified times and indices.
    """
    ...


def update_battery_state(
        python_energy_or_current_array: np.ndarray,
        time_step: float,
        initial_state_of_charge: float,
        initial_polarization_potential: float,
        python_internal_resistance_lookup: np.ndarray,
        python_open_circuit_voltage_lookup: np.ndarray,
        python_polarization_resistance_lookup: np.ndarray,
        python_polarization_capacitance_lookup: np.ndarray,
        nominal_charge_capacity: float,
        is_power: bool,
        quantization_step: float,
        min_soc: float
) -> tuple[np.ndarray, np.ndarray]:
    """
    Updates the battery state (SOC and terminal voltage) based on energy/current input.

    :param python_energy_or_current_array: Array of energy or current input values (f64).
    :param time_step: Time step (f64).
    :param initial_state_of_charge: Initial state of charge (f64).
    :param initial_polarization_potential: Initial polarization potential (f64).
    :param python_internal_resistance_lookup: Array of internal resistance values (f64).
    :param python_open_circuit_voltage_lookup: Array of open-circuit voltage values (f64).
    :param python_polarization_resistance_lookup: Array of polarization resistance values (f64).
    :param python_polarization_capacitance_lookup: Array of polarization capacitance values (f64).
    :param nominal_charge_capacity: Nominal charge capacity (f64).
    :param is_power: Boolean flag to indicate if the input is power (`True`) or current (`False`).
    :param quantization_step: The step size used to quantize the SOC (f64).
    :param min_soc: The minimum SOC used when computing the lookup tables

    :return: A tuple containing:
        - An array of updated SOC values (f64).
        - An array of updated terminal voltage values (f64).
    """
    ...



from typing import Optional
import numpy as np
from abc import ABC, abstractmethod


class BaseMeteorology(ABC):
    def __init__(self):
        self._wind_speed: Optional[np.ndarray] = None
        self._wind_direction: Optional[np.ndarray] = None
        self._solar_irradiance: Optional[np.ndarray] = None
        self._weather_indices: Optional[np.ndarray] = None

    def _return_if_available(self, attr):
        if (value := getattr(self, attr)) is not None:
            return value
        else:
            raise UnboundLocalError(f"{attr} is not available!")

    @property
    def wind_speed(self) -> np.ndarray:
        """
        Return the wind speeds in m/s at every tick, if available.

        :return: ``ndarray`` of wind speeds in m/s at every tick
        :raises UnboundLocalError: if wind speeds are not available.
        """
        return self._return_if_available("_wind_speed")

    @property
    def wind_direction(self) -> np.ndarray:
        """
        Return the wind direction in degrees, following the meteorological convention, if available.

        :return: ``ndarray`` of wind directions in degrees at every tick.
        :raises UnboundLocalError: if wind directions are not available.
        """
        return self._return_if_available("_wind_direction")

    @property
    def solar_irradiance(self) -> np.ndarray:
        """
        Return the solar irradiance in W/m^2 every tick, if available.

        :return: ``ndarray`` of solar irradiances in W/m^2 at every tick
        :raises UnboundLocalError: if solar irradiances are not available.
        """
        return self._return_if_available("_solar_irradiance")

    @property
    def weather_indices(self) -> np.ndarray:
        """
        Return the weather indices at every tick, if available.

        :return: ``ndarray`` of weather indices at every tick
        :raises UnboundLocalError: if weather indices are not available.
        """
        return self._return_if_available("_weather_indices")

    @abstractmethod
    def spatially_localize(self, cumulative_distances: np.ndarray) -> None:
        raise NotImplementedError

    @abstractmethod
    def temporally_localize(self, unix_timestamps, start_time, tick) -> None:
        raise NotImplementedError

    @abstractmethod
    def calculate_solar_irradiances(self, coords, time_zones, local_times, elevations):
        raise NotImplementedError



import numpy as np
import physics_rs
from typing import Callable, TypeAlias, Protocol, runtime_checkable, Optional, cast
from numpy.typing import NDArray


SOCDependent: TypeAlias = Callable[[float | NDArray[float]], float | NDArray[float]]


@runtime_checkable
class EquivalentCircuitModelConfig(Protocol):
    """
    A specification for a configuration object which contains the requisite data to specify
    a `EquivalentCircuitBatteryModel`.
    """

    @property
    def get_Uoc(self) -> SOCDependent:
        """
        A map from an SOC to Uoc (open-circuit voltage).
        Should be compatible with non-vectorized and vectorized calls: float -> float or NDArray -> NDArray
        """
        ...

    @property
    def get_R_0(self) -> SOCDependent:
        """
        A map from an SOC to R_0 (internal resistance).
        Should be compatible with non-vectorized and vectorized calls: float -> float or NDArray -> NDArray
        """
        ...

    @property
    def get_R_P(self) -> SOCDependent:
        """
        A map from an SOC to R_P (polarization resistance).
        Should be compatible with non-vectorized and vectorized calls: float -> float or NDArray -> NDArray
        """
        ...

    @property
    def get_C_P(self) -> SOCDependent:
        """
        A map from an SOC to C_P (polarization capacitance).
        Should be compatible with non-vectorized and vectorized calls: float -> float or NDArray -> NDArray
        """
        ...

    @property
    def Q_total(self) -> float:
        """
        The total charge capacity of the battery pack, in Coulombs.
        """
        ...


class EquivalentCircuitBatteryModel:
    """
    A first-order Thevenin equivalent model of a lithium-ion battery pack
    """

    def __init__(self, battery_config: EquivalentCircuitModelConfig, state_of_charge: float = 1.0):
        """
        Constructor for the EquivalentCircuitBatteryModel class.

        :param BatteryModelConfig battery_config: Configuration object containing the battery's parameters and data.
        :param float state_of_charge: Initial state of charge of the battery (default is 1.0, fully charged).
        """

        # We initialize the active components as uncharged
        self._U_P = 0.0  # V
        self._U_L = 0.0  # V
        self._state_of_charge = state_of_charge
        self._nominal_charge_capacity = battery_config.Q_total

        # Now, the config contains methods to map SOC to each respective parameter.
        # We can't efficiently pass these functions to compiled libraries.
        # Instead, we will pre-compute the parameters as a function of SOC
        # to create fine lookup tables as a portable substitute for runtime computation.

        # Things are going to get a tiny bit messy here, so we will go through this carefully.
        # I'll write what each resulting map achieves below each code block.

        # Firstly, we're going to discretize SOC by making a range of SOC values in the range [-0.05, 1.1], because
        # sometimes we are marginally outside the range (0.0, 1.0].
        # We will quantize at about 4 digits of precision, so ~10,000 values
        self._min_soc = -0.05
        self._max_soc = 1.1
        self._num_indices = int((self._max_soc - self._min_soc) * 10000)
        SOC_values = np.linspace(self._min_soc, self._max_soc, self._num_indices, dtype=float)
        # maps: (discrete index) -> (SOC)

        # Now, we're going to create a map from an arbitrary SOC, to the index of the closest SOC
        # value in our quantized SOC range (`SOC_values`)
        self._quantization_step: float = (self._max_soc - self._min_soc) / self._num_indices
        self._soc_to_index = lambda _soc: int(
            max(0, min(self._num_indices - 1, (_soc - self._min_soc) // self._quantization_step))
        )
        # maps: (SOC) -> (discrete index)

        # Now, calculate the value of each parameter for each discrete SOC value using the injected `get_` functions
        self._U_oc_lookup: NDArray[float] = battery_config.get_Uoc(SOC_values)
        self._R_0_lookup: NDArray[float] = battery_config.get_R_0(SOC_values)
        self._R_P_lookup: NDArray[float] = battery_config.get_R_P(SOC_values)
        self._C_P_lookup: NDArray[float] = battery_config.get_C_P(SOC_values)
        # maps: (discrete index) -> (parameter)

        # Finally, combine the above maps to create a map from an arbitrary SOC to each battery parameter, using
        # the discrete lookup tables
        # These `cast` calls just promise to the type-checker that these will map floats to floats
        self._U_oc = cast(Callable[[float], float], lambda SOC: self._U_oc_lookup[self._soc_to_index(SOC)])
        self._R_0 = cast(Callable[[float], float], lambda SOC: self._R_0_lookup[self._soc_to_index(SOC)])
        self._R_P = cast(Callable[[float], float], lambda SOC: self._R_P_lookup[self._soc_to_index(SOC)])
        self._C_P = cast(Callable[[float], float], lambda SOC: self._C_P_lookup[self._soc_to_index(SOC)])
        # maps: ((SOC) -> (discrete index)) -> ((discrete index) -> (parameter))  |==>  (SOC) -> (parameter)

        self._tau: Callable[[float], float] = lambda soc: self._R_P(soc) * self._C_P(soc)  # Characteristic Time in s

    def update_array(
            self,
            tick: float,
            delta_energy_array: Optional[NDArray] = None,
            current_array: Optional[NDArray] = None,
            use_compiled: bool = True
    ) -> tuple[NDArray, NDArray]:
        """
        Compute the battery's state of charge and terminal voltage over time in response to a
        time series of energy/current draw from a load.

        Only ONE of `current_array` or `delta_energy_array` should be provided.

        Notes
        -----
        If both current and power are known, current should be provided.
        The model implementation requires current for calculations, so it must be derived from power if power
        was provided.
        Computing current from power relies on voltage, which is a model output, and therefore
        the derived current could be less accurate.

        :param NDArray delta_energy_array: Array of energy changes (J) at each time step.
        :param float tick: Time interval for each step (seconds).
        :param NDArray current_array: Array of current draw (positive sign convention) in Amperes at each time step.
        :param bool use_compiled: If `True`, use compiled binaries for calculations.
            Disable for better debugging.
        :return: A tuple containing arrays for state-of-charge and terminal voltage.
        :raises ValueError: If BOTH or NEITHER of `current_array` or `delta_energy_array` are provided.
        :rtype: tuple[NDArray, NDArray]
        """
        if (delta_energy_array is None) == (current_array is None):  # Enforce that only one should be provided
            raise ValueError("Exactly one of `delta_energy_array` or `current_array` "
                             "must be provided, not both or neither.")

        energy_or_current = delta_energy_array if delta_energy_array is not None else current_array

        if use_compiled:
            return physics_rs.update_battery_state(
                energy_or_current,
                tick,
                self._state_of_charge,
                self._U_P,
                self._R_0_lookup,
                self._U_oc_lookup,
                self._R_P_lookup,
                self._C_P_lookup,
                self._nominal_charge_capacity,
                current_array is None,  # Pass to the library if `energy_or_current` is current or power,
                self._quantization_step,
                self._min_soc
            )

        else:
            return self._update_array_py(energy_or_current, tick, current_array is None)

    def _update_array_py(self, energy_or_current, tick, is_power):
        """
        Perform energy calculations using Python (fallback method if Rust is disabled).

        :param NDArray energy_or_current: Array of energy changes (J) at each time step.
        :param float tick: Time interval for each step (seconds).

        :return: A tuple containing arrays for state-of-charge and voltage.
        """
        soc = np.empty_like(energy_or_current, dtype=float)
        voltage = np.empty_like(energy_or_current, dtype=float)

        for (i, value) in enumerate(energy_or_current):
            if is_power:
                # Use the last voltage to calculate current, or an absurdly large number if it is the first,
                # because we don't know voltage yet.
                # We will have a very small initial current, no matter what.
                # We shouldn't be starting to simulate when the battery is in an active state anyway,
                # so this should be an alright compromise.
                last_terminal_voltage = voltage[i - 1] if i - 1 >= 0 else 10000

                current: float = value / (tick * last_terminal_voltage)
            else:
                current = value

            self._evolve(current, tick)
            soc[i] = self._state_of_charge
            voltage[i] = self._U_L

        return soc, voltage

    def _evolve(self, current: float, tick: float) -> None:
        """
        Update the battery state given the current and time elapsed.

        :param float current: Current applied to the battery (A).
            Positive for charging, negative for discharging.
        :param float tick: Time interval over which the power is applied (seconds).
        """
        soc = self._state_of_charge  # State of Charge (dimensionless, 0 < soc < 1)
        U_P = self._U_P  # Polarization Potential (V)
        R_P = self._R_P(soc)  # Polarization Resistance (Ohms)
        U_oc = self._U_oc(soc)  # Open-Circuit Potential (V)
        R_0 = self._R_0(soc)  # Ohmic Resistance (Ohms)
        Q = self._nominal_charge_capacity  # Nominal Charge Capacity (C)
        tau = self._tau(soc)  # Time constant (s)

        new_soc = soc + (current * tick / Q)
        new_U_P = np.exp(-tick / tau) * U_P + current * R_P * (1 - np.exp(-tick / tau))

        self._state_of_charge = new_soc
        self._U_P = new_U_P
        self._U_L = U_oc + new_U_P + (current * R_0)



import math
import numpy as np
from numpy.typing import NDArray
from physics.models.motor.base_motor import BaseMotor
from physics.models.constants import ACCELERATION_G, AIR_DENSITY


class BasicMotor(BaseMotor):
    def __init__(self, vehicle_mass, road_friction, tire_radius, vehicle_frontal_area, drag_coefficient):
        super().__init__()

        # Instantaneous voltage supplied by the battery to the motor controller
        self.dc_v = 0

        # Instantaneous current supplied by the battery to the motor controller
        self.dc_i = 0

        self.input_power = 0
        self.vehicle_mass = vehicle_mass
        self.acceleration_g = ACCELERATION_G
        self.road_friction = road_friction
        self.tire_radius = tire_radius

        self.air_density = AIR_DENSITY
        self.vehicle_frontal_area = vehicle_frontal_area
        self.drag_coefficient = drag_coefficient

        self.friction_force = (self.vehicle_mass * self.acceleration_g * self.road_friction)

        self.e_mc = 0.98  # motor controller efficiency, subject to change
        self.e_m = 0.9  # motor efficiency, subject to change

        # print("torque experienced by motor: {} Nm".format(self.constant_torque))

    @staticmethod
    def calculate_motor_efficiency(motor_angular_speed, motor_output_energy, tick, *args, **kwargs):
        """

        Calculates a NumPy array of motor efficiency from NumPy array of operating angular speeds and NumPy array
            of output power. Based on data obtained from NGM SC-M150 Datasheet and modelling done in MATLAB

        r squared value: 0.873

        :param np.ndarray motor_angular_speed: (float[N]) angular speed motor operates in rad/s
        :param np.ndarray motor_output_energy: (float[N]) energy motor outputs to the wheel in J
        :param float tick: length of 1 update cycle in seconds
        :returns e_m: (float[N]) efficiency of the motor
        :rtype: np.ndarray

        """

        # Power = Energy / Time
        motor_output_power = motor_output_energy * tick
        rads_rpm_conversion_factor = 30 / math.pi

        revolutions_per_minute = motor_angular_speed * rads_rpm_conversion_factor

        e_m = calculate_motor_efficiency(motor_output_power, revolutions_per_minute)

        e_m[e_m < 0.7382] = 0.7382
        e_m[e_m > 1] = 1

        return e_m

    @staticmethod
    def calculate_motor_controller_efficiency(motor_angular_speed, motor_output_energy, tick):
        """

        Calculates a NumPy array of motor controller efficiency from NumPy array of operating angular speeds and
        NumPy array of output power. Based on data obtained from the WaveSculptor Motor Controller Datasheet efficiency
        curve for a 90 V DC Bus and modelling done in MATLAB.

        r squared value: 0.7431

        :param np.ndarray motor_angular_speed: (float[N]) angular speed motor operates in rad/s
        :param np.ndarray motor_output_energy: (float[N]) energy motor outputs to the wheel in J
        :param float tick: length of 1 update cycle in seconds
        :returns e_mc (float[N]) efficiency of the motor controller
        :rtype: np.ndarray

        """

        # Ignore nan warning. Set nan value to 0
        np.seterr(divide='ignore', invalid='ignore')

        # Power = Energy / Time
        motor_output_power = motor_output_energy / tick

        # Torque = Power / Angular Speed
        motor_torque_array = np.nan_to_num(motor_output_power / motor_angular_speed)

        np.seterr(divide='warn', invalid='warn')

        e_mc = calculate_motor_controller_efficiency(motor_angular_speed, motor_torque_array)

        e_mc[e_mc < 0.9] = 0.9
        e_mc[e_mc > 1] = 1

        return e_mc

    def calculate_net_force(self,
                            required_speed_kmh: NDArray,
                            wind_speeds: NDArray,
                            gradients: NDArray
                            ) -> tuple[NDArray, NDArray]:
        """
        Calculate the net force on the car, and the required wheel angular velocity.
        Currently, considers:
            1. Rolling resistance of the wheels on the road
            2. Drag force (wind + forward velocity)
            3. Acceleration force (a = F / m)
            4. Gravitational force (force to go uphill)

        :return: net force in N, wheel angular velocity in rad/s
        """
        required_speed_ms = required_speed_kmh / 3.6

        acceleration_ms2 = np.clip(np.gradient(required_speed_ms), a_min=0, a_max=None)
        acceleration_force = acceleration_ms2 * self.vehicle_mass

        required_angular_speed_rads = required_speed_ms / self.tire_radius

        drag_forces = 0.5 * self.air_density * (
                (required_speed_ms + wind_speeds) ** 2) * self.drag_coefficient * self.vehicle_frontal_area

        angles = np.arctan(gradients)
        g_forces = self.vehicle_mass * self.acceleration_g * np.sin(angles)

        road_friction_array = self.road_friction * self.vehicle_mass * self.acceleration_g * np.cos(angles)

        net_force = road_friction_array + drag_forces + g_forces + acceleration_force

        return net_force, required_angular_speed_rads

    def calculate_energy_in(self, required_speed_kmh, gradients, wind_speeds, tick, **kwargs):
        """

        Create a function which takes in array of elevation, array of wind speed, required
            speed, returns the consumed energy.

        :param np.ndarray required_speed_kmh: (float[N]) required speed array in km/h
        :param np.ndarray gradients: (float[N]) gradient at parts of the road
        :param np.ndarray wind_speeds: (float[N]) speeds of wind in m/s, where > 0 means against the direction of the vehicle
        :param float tick: length of 1 update cycle in seconds
        :returns: (float[N]) energy expended by the motor at every tick
        :rtype: np.ndarray

        """
        net_force, required_angular_speed_rads = self.calculate_net_force(required_speed_kmh, wind_speeds, gradients)

        motor_output_energies = required_angular_speed_rads * net_force * self.tire_radius * tick
        motor_output_energies = np.clip(motor_output_energies, a_min=0, a_max=None)

        e_m = self.calculate_motor_efficiency(required_angular_speed_rads, motor_output_energies, tick)
        e_mc = self.calculate_motor_controller_efficiency(required_angular_speed_rads, motor_output_energies, tick)

        motor_controller_input_energies = motor_output_energies / (e_m * e_mc)

        # Filter out and replace negative energy consumption as 0
        motor_controller_input_energies = np.where(motor_controller_input_energies > 0,
                                                   motor_controller_input_energies, 0)

        return motor_controller_input_energies

    def __str__(self):
        return (f"Tire radius: {self.tire_radius}m\n"
                f"Rolling resistance coefficient: {self.road_friction}\n"
                f"Vehicle mass: {self.vehicle_mass}kg\n"
                f"Acceleration of gravity: {self.acceleration_g}m/s^2\n"
                f"Motor controller efficiency: {self.e_mc}%\n"
                f"Motor efficiency: {self.e_m}%\n")


def calculate_motor_efficiency(motor_output_power, revolutions_per_minute):
    return 0.7382 - (6.281e-5 * motor_output_power) + (6.708e-4 * revolutions_per_minute) \
        - (2.89e-8 * motor_output_power ** 2) + (2.416e-7 * motor_output_power * revolutions_per_minute) \
        - (8.672e-7 * revolutions_per_minute ** 2) + (5.653e-12 * motor_output_power ** 3) \
        - (1.74e-11 * motor_output_power ** 2 * revolutions_per_minute) \
        - (7.322e-11 * motor_output_power * revolutions_per_minute ** 2) \
        + (3.263e-10 * revolutions_per_minute ** 3)


def calculate_motor_controller_efficiency(motor_angular_speed, motor_torque_array):
    return 0.7694 + (0.007818 * motor_angular_speed) + (0.007043 * motor_torque_array) \
        - (1.658e-4 * motor_angular_speed ** 2) - (1.806e-5 * motor_torque_array * motor_angular_speed) \
        - (1.909e-4 * motor_torque_array ** 2) + (1.602e-6 * motor_angular_speed ** 3) \
        + (4.236e-7 * motor_angular_speed ** 2 * motor_torque_array) \
        - (2.306e-7 * motor_angular_speed * motor_torque_array ** 2) \
        + (2.122e-06 * motor_torque_array ** 3) - (5.701e-09 * motor_angular_speed ** 4) \
        - (2.054e-9 * motor_angular_speed ** 3 * motor_torque_array) \
        - (3.126e-10 * motor_angular_speed ** 2 * motor_torque_array ** 2) \
        + (1.708e-09 * motor_angular_speed * motor_torque_array ** 3) \
        - (8.094e-09 * motor_torque_array ** 4)



import numpy as np
from filterpy.kalman import ExtendedKalmanFilter
from typing import Protocol, runtime_checkable, cast, Callable
from physics.models.battery import EquivalentCircuitModelConfig
from numpy.typing import NDArray


@runtime_checkable
class FilteredBatteryModelConfig(Protocol):
    """
    A specification for a configuration object which contains the requisite data to specify
    a `FilteredBatteryModel`.
    """
    @property
    def battery_model_config(self) -> EquivalentCircuitModelConfig:
        """
        Configuration of the underlying `EquivalentCircuitModel`.
        """
        ...

    @property
    def process_noise_matrix(self) -> NDArray[float]:
        """
        A 2x2 matrix containing the process noise covariance matrix where [0, 0] is the SOC evolution
        noise and [1, 1] is the polarization potential evolution noise.
        """
        ...

    @property
    def state_covariance_matrix(self) -> NDArray[float]:
        """
        A 2x2 matrix containing the state covariance matrix where [0, 0] is the SOC covariance
        noise and [1, 1] is the polarization potential covariance.
        """
        ...

    @property
    def measurement_noise_vector(self) -> NDArray[float]:
        """
        A 1x1 vector containing the noise expected in the terminal voltage measurement.
        """
        ...


class FilteredBatteryModel:
    """
    `FilteredBatteryModel` is a first-order Thevenin equivalent model of a lithium-ion battery packed, wrapped
    in a Kalman filter which uses voltage measurements with model predictions.
    """
    def __init__(
            self,
            battery_config: FilteredBatteryModelConfig,
            initial_SOC: float = 1.0,
            initial_Uc: float = 0.0,
            alpha: float = 0.9
    ):
        """
        :param FilteredBatteryModelConfig battery_config: Contains Kalman filter state estimation configuration and
            underlying equivalent circuit model configuration.
        :param float initial_SOC: Initial SOC of the battery, in the range (0, 1].
        :param float initial_Uc: Initial polarization voltage of the battery in Volts.
        """
        # Initial state
        assert 0.0 <= initial_SOC <= 1.1, "`initial_SOC` must be in (0, 1.1]!"

        self._SOC = initial_SOC     # State of Charge
        self._Uc = initial_Uc       # Polarization Voltage

        # Load Config data
        self._Q_total = battery_config.battery_model_config.Q_total

        # These `cast` calls just promise to the type-checker that these will map floats to floats
        self._U_oc = cast(Callable[[float], float], battery_config.battery_model_config.get_Uoc)
        self._R_0 = cast(Callable[[float], float], battery_config.battery_model_config.get_R_0)
        self._R_P = cast(Callable[[float], float], battery_config.battery_model_config.get_R_P)
        self._C_P = cast(Callable[[float], float], battery_config.battery_model_config.get_C_P)

        self._tau: Callable[[float], float] = lambda soc: self._R_P(soc) * self._C_P(soc)  # Characteristic Time in s

        def central_difference_derivative(func, value, h=1e-6):
            """
            Compute the derivative of an arbitrary function `func` at `SOC` using central difference.

            :param func: The function to differentiate.
            :param value: The point at which to compute the derivative.
            :param h: Step size for the finite difference.
            :return: The derivative of the function at `SOC`.
            """
            return (func(value + h) - func(value - h)) / (2 * h)

        self._dUoc_dSOC = lambda soc: central_difference_derivative(self._U_oc, np.minimum(1.0, soc))  # dUOC wrt to SOC
        self._dR0_dSOC = lambda soc: central_difference_derivative(self._R_0, np.minimum(1.0, soc))    # dR0 wrt to SOC

        # Initializing the EKF object
        self._ekf = ExtendedKalmanFilter(dim_x=2, dim_z=1)

        # State Vector
        self._ekf.x = np.array([
            self._SOC,
            self._Uc]
        )

        self._ekf.P = battery_config.state_covariance_matrix
        self._ekf.Q = battery_config.process_noise_matrix
        self._ekf.R = battery_config.measurement_noise_vector

        assert 0 <= alpha <= 1, "`alpha` should be between 0 and 1!"
        self._alpha = alpha

        self._filtered_I = 0
        self._predicted_measurement = 0

    @property
    def SOC(self) -> float:
        """
        Return the current SOC of the battery.

        :return: The current state of charge.
        """
        return self._SOC

    @property
    def Uc(self) -> float:
        """
        Return the polarization voltage of the battery.

        :return: The current polarization voltage.
        """
        return self._Uc

    @property
    def Ut(self) -> float:
        """
        Return the predicted terminal voltage for the last prediction step.

        :return: The predicted terminal voltage.
        """
        return self._predicted_measurement

    def update_filter(self, measured_Ut, current):
        """
        Update the filter based on a new measurement and the predicted state.
        This function should be called after `predict_state` in a typical predict-update workflow.

        :param float measured_Ut: The actual voltage across the terminals of the battery.
        :param float current: The current being sourced by the battery.
        """
        # Simple low-pass filter to current
        self._filtered_I = self._alpha * self._filtered_I + (1 - self._alpha) * current

        self._ekf.update(z=measured_Ut, HJacobian=self._measurement_jacobian, Hx=self._measurement_function)

        self._SOC, self._Uc = self._ekf.x
        self._SOC = np.clip(self._SOC, 0.0, 1.1)

    def predict_state(self, current, time_step):
        """
        Predict the next evolution of the state vector (SOC, Uc).
        This function should be called before updating the filter in a typical predict-update workflow.

        :param float current: The current being sourced by the battery.
            Sign convention is that positive indicates current being drawn.
        :param float time_step: Time elapsed between this prediction and the last updated state of the filter (seconds).
        """
        # Control matrix B (for input current I_k)
        self._ekf.B = np.array([
            -time_step / self._Q_total,
            self._R_P(self._SOC) * (1 - np.exp(-time_step / (self._tau(self._SOC)))),
        ])
        self._ekf.F = self._state_jacobian(time_step)

        self._ekf.predict(u=current)
        self._SOC, self._Uc = self._ekf.x

    def predict_then_update(self, measured_Ut: float, current: float, time_step: float):
        """
        Predict the next evolution of the state vector (SOC, Uc), then update the filter
        based on this prediction and a measurement. Abstracts the full predict-update workflow of the EKF.

        :param float measured_Ut: The actual voltage across the terminals of the battery.
        :param float current: The current being sourced by the battery. Positive indicates current being drawn.
        :param float time_step: Time elapsed between this prediction and the last updated state of the filter (seconds).
        """
        self.predict_state(current, time_step)
        self.update_filter(measured_Ut, current)

    def _state_jacobian(self, time_step):
        """
        Return the state Jacobian matrix for the current time step.

        :param float time_step: Time elapsed between this prediction and the last updated state of the filter (seconds).
        :return: The state Jacobian matrix.
        :rtype: np.ndarray
        """
        return np.array([[1, 0], [0, np.exp(-time_step / self._tau(self._SOC))]])

    def _measurement_jacobian(self, x):
        """
        Return the measurement Jacobian matrix for the current state vector.

        :param list[float, float] x: The state vector [SOC, Uc].
        :return: The measurement Jacobian matrix.
        :rtype: np.ndarray
        """
        SOC = x[0]
        dUoc_dSOC = self._dUoc_dSOC(SOC)
        dR0_dSOC = self._dR0_dSOC(SOC)

        return np.array([[dUoc_dSOC - dR0_dSOC * self._filtered_I, -1]])

    def _measurement_function(self, x) -> float:
        """
        Return the measurement function relating terminal voltage to SOC and polarization voltage.

        :param list[float, float] x: The state vector [SOC, Uc].
        :return: The predicted terminal voltage.
        """
        SOC, Uc = x
        Uoc = self._U_oc(SOC)
        R0 = self._R_0(SOC)
        self._predicted_measurement = Uoc - Uc - R0 * self._filtered_I

        return self._predicted_measurement



from .base_battery import BaseBattery
from .basic_battery import BasicBattery
from .battery_model import EquivalentCircuitBatteryModel, EquivalentCircuitModelConfig, SOCDependent
from .kalman_filter import FilteredBatteryModel, FilteredBatteryModelConfig
from .battery_config import BatteryModelConfig, load_battery_config, KalmanFilterConfig

__all__ = [
    "BaseBattery",
    "BasicBattery",
    "EquivalentCircuitBatteryModel",
    "FilteredBatteryModel",
    "BatteryModelConfig",
    "load_battery_config",
    "EquivalentCircuitModelConfig",
    "FilteredBatteryModelConfig",
    "KalmanFilterConfig",
    "SOCDependent"
]



import os
import sys
sys.path.insert(0, os.path.abspath('../../'))  # Adjust as necessary
print(sys.path)
print(os.listdir(os.getcwd()))
print(os.listdir(sys.path[0]))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'UBC Solar Physics'
copyright = '2024, UBC Solar'
author = 'Joshua Riefman'
release = "1.0.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',   # For Google/NumPy style docstrings
    'sphinx.ext.autosummary',  # To generate summary tables for modules
    'myst_parser',           # For Markdown support if needed
]

html_theme = "pydata_sphinx_theme"

autodoc_mock_imports = ['core']
source_suffix = ['.rst', '.md']

templates_path = ['_templates']
exclude_patterns = []
autosummary_generate = True

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ['_static']



import numpy as np
from numpy.polynomial import Polynomial
from physics.models.battery.base_battery import BaseBattery


class BasicBattery(BaseBattery):
    """
    Class representing the DayBreak battery pack.

    Attributes:
        max_voltage (float): maximum voltage of the DayBreak battery pack (V)
        min_voltage (float): minimum voltage of the DayBreak battery pack (V)
        max_current_capacity (float): nominal capacity of the DayBreak battery pack (Ah)
        max_energy_capacity (float): nominal energy capacity of the DayBreak battery pack (Wh)

        state_of_charge (float): instantaneous battery state-of-charge (0.00 - 1.00)
        discharge_capacity (float): instantaneous amount of charge extracted from battery (Ah)
        voltage (float): instantaneous voltage of the battery (V)
        stored_energy (float): instantaneous energy stored in the battery (Wh)
    """

    def __init__(self, state_of_charge, max_voltage, min_voltage, max_current_capacity, max_energy_capacity):
        """

        Constructor for BasicBattery class.

        :param float state_of_charge: initial battery state of charge

        """

        # ----- DayBreak battery constants -----

        self.max_voltage = max_voltage
        self.min_voltage = min_voltage
        self.max_current_capacity = max_current_capacity
        self.max_energy_capacity = max_energy_capacity

        # ----- DayBreak battery equations -----

        self.calculate_voltage_from_discharge_capacity = calculate_voltage_from_discharge_capacity()

        self.calculate_energy_from_discharge_capacity = calculate_energy_from_discharge_capacity()

        self.calculate_soc_from_discharge_capacity = calculate_soc_from_discharge_capacity(self.max_current_capacity)

        self.calculate_discharge_capacity_from_soc = calculate_discharge_capacity_from_soc(self.max_current_capacity)

        self.calculate_discharge_capacity_from_energy = calculate_discharge_capacity_from_energy()

        # ----- DayBreak battery variables -----

        self.state_of_charge = state_of_charge

        # SOC -> discharge_capacity
        self.discharge_capacity = self.calculate_discharge_capacity_from_soc(self.state_of_charge)

        # discharge_capacity -> voltage
        self.voltage = self.calculate_voltage_from_discharge_capacity(self.discharge_capacity)

        # discharge_capacity -> energy
        self.stored_energy = self.max_energy_capacity - self.calculate_energy_from_discharge_capacity(
            self.discharge_capacity)

        # ----- DayBreak battery initialisation -----

        super().__init__(self.stored_energy, self.max_current_capacity, self.max_energy_capacity,
                         self.max_voltage, self.min_voltage, self.voltage, self.state_of_charge)

    def update_array(self, cumulative_energy_array):
        """
        Performs energy calculations with NumPy arrays

        :param cumulative_energy_array: a NumPy array containing the cumulative energy changes at each time step
        experienced by the battery

        :return: soc_array – a NumPy array containing the battery state of charge at each time step

        :return: voltage_array – a NumPy array containing the voltage of the battery at each time step

        :return: stored_energy_array– a NumPy array containing the energy stored in the battery at each time step

        """

        stored_energy_array = np.full_like(cumulative_energy_array, fill_value=self.stored_energy)
        stored_energy_array += cumulative_energy_array / 3600
        stored_energy_array = np.clip(stored_energy_array, a_min=0, a_max=self.max_energy_capacity)

        energy_discharged_array = np.full_like(cumulative_energy_array, fill_value=self.max_energy_capacity) - \
                                  stored_energy_array

        discharge_capacity_array = self.calculate_discharge_capacity_from_energy(energy_discharged_array)

        soc_array = self.calculate_soc_from_discharge_capacity(discharge_capacity_array)
        voltage_array = self.calculate_voltage_from_discharge_capacity(discharge_capacity_array)

        return soc_array, voltage_array, stored_energy_array

    def get_raw_soc(self, cumulative_energy_array):
        """

        Return the not truncated (SOC is allowed to go above 100% and below 0%) state of charge.

        :param np.ndarray cumulative_energy_array: a NumPy array containing the cumulative energy changes at each time step
        experienced by the battery

        :return: a NumPy array containing the battery state of charge at each time step
        :rtype: np.ndarray

        """

        stored_energy_array = np.full_like(cumulative_energy_array, fill_value=self.stored_energy)
        stored_energy_array += cumulative_energy_array / 3600

        energy_discharged_array = np.full_like(cumulative_energy_array, fill_value=self.max_energy_capacity) - stored_energy_array

        discharge_capacity_array = self.calculate_discharge_capacity_from_energy(energy_discharged_array)

        soc_array = self.calculate_soc_from_discharge_capacity(discharge_capacity_array)

        return soc_array


def calculate_voltage_from_discharge_capacity():
    return Polynomial([117.6, -0.858896])  # -0.8589x + 117.6


def calculate_energy_from_discharge_capacity():
    return Polynomial([0, 117.6, -0.429448])  # -0.4294x^2 + 117.6x


def calculate_soc_from_discharge_capacity(max_current_capacity):
    return Polynomial([1, -1 / max_current_capacity])


def calculate_discharge_capacity_from_soc(max_current_capacity):
    return Polynomial([max_current_capacity, -max_current_capacity])


def calculate_discharge_capacity_from_energy():
    return lambda x: 136.92 - np.sqrt(18747.06027 - 2.32857 * x)



from data_tools.collections.time_series import TimeSeries
from data_tools.query.influxdb_query import DBClient
from datetime import datetime, timezone
import numpy as np
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
import pandas as pd


client = DBClient()

# ISO 8601-compliant times corresponding to pre-competition testing


# Day 1
start_dt = datetime(2024, 7, 16, 14, 23, 57, tzinfo=timezone.utc)
stop_dt = datetime(2024, 7, 16, 21, 34, 15, tzinfo=timezone.utc)



client = DBClient()


# We can, in one line, make a query to InfluxDB and parse 
# the data into a powerful format: the `TimeSeries` class.
pack_current_data: TimeSeries = client.query_time_series(
    start=start_dt,
    stop=stop_dt,
    field="PackCurrent",
)


battery_voltage_data: TimeSeries = client.query_time_series(
    start=start_dt,
    stop=stop_dt,
    field="TotalPackVoltage",
)

voltage_aligned, current_aligned = TimeSeries.align(battery_voltage_data, pack_current_data)


def timeseries_to_csv(timeseries_instance, output_file):
    # Extract time (as datetime) and values
    times = timeseries_instance.datetime_x_axis
    values = timeseries_instance
    
    # Create a DataFrame with the time and value columns
    df = pd.DataFrame({
        'Time': times,
        'Value': values
    })
    
    # Write the DataFrame to a CSV file
    df.to_csv(output_file, index=False)


timeseries_to_csv(voltage_aligned, 'voltage.csv')
timeseries_to_csv(current_aligned, 'current.csv')





from physics.models.regen import BaseRegen
import numpy as np


class BasicRegen(BaseRegen):
    GRAVITY = 9.81
    EFFICIENCY = 0.5  # currently set to 50% but best case scenario is 60-70%

    def __init__(self, vehicle_mass):
        super().__init__()
        self.min_decel_mag = 0
        self.vehicle_mass = vehicle_mass
        self.kmh_to_mps = 0.278

    def get_regen_efficiency(self, speed_array):
        """
        Returns a numpy array of regen efficiency percentage based on the vehicle speed in m/s.

        :param speed_array: a numpy array of speeds in m/s
        :returns: numpy array of regen efficiency percentage
        """
        # Efficiency polynomial, more details can be found in regen_analysis folder located in data_analysis
        efficiency_poly = [0.022288416685942, 0.026545396753597]

        return np.polyval(efficiency_poly, speed_array)

    def calculate_produced_energy(self, speed_kmh, gis_route_elevations, min_regen_speed, max_power):
        """
        Returns a numpy array containing the energy produced by regen
        during each tick of the race based on the change in energy in that tick
        :param speed_kmh: an array containing the speeds at each tick
        :param gis_route_elevations: an array containing elevations on the route at each tick
        """
        # get the changes of energy from tick i to tick i + 1
        speed_ms = speed_kmh / 3.6  # Convert to m/s from km/h
        delta_kinetic_energy = np.diff((1 / 2) * self.vehicle_mass * pow(speed_ms, 2), append=[0])
        delta_potential_energy = np.diff(self.vehicle_mass * self.GRAVITY * gis_route_elevations, append=[0])

        # get the total change in energy at each tick
        delta_energy = delta_kinetic_energy + delta_potential_energy

        # create regen energy produced array
        # if delta_energy is negative, we regen that energy back at the set efficiency rate; else 0 energy regen
        efficiencies = self.get_regen_efficiency(speed_ms)
        produced_energy = np.where(delta_energy < 0, abs(delta_energy) * efficiencies, 0)

        # Regen does not occur below a certain speed
        produced_energy = np.where(speed_ms >= min_regen_speed, produced_energy, 0)

        # Regen power is capped by current limitations

        return np.clip(produced_energy, a_min=0, a_max=max_power)



from setuptools_scm import get_version
from pathlib import Path

# Get the version from setuptools_scm
version = get_version()

# Path to the version file specified in your pyproject.toml
version_file = Path('physics/_version.py')

# Write the version into _version.py
version_file.write_text(f"__version__ = '{version}'\n")



import numpy as np
from haversine import haversine, Unit
from numpy.typing import NDArray
from physics.models.motor import BasicMotor


class AdvancedMotor(BasicMotor):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.cornering_coefficient = 15  # tuned to Day 1 and 3 FSGP data

    def calculate_energy_in(self, required_speed_kmh, gradients, wind_speeds, tick, coords):
        """
        A function which takes in array of elevation, array of wind speed, required
            speed, returns the consumed energy.

        :param np.ndarray required_speed_kmh: (float[N]) required speed array in km/h
        :param np.ndarray gradients: (float[N]) gradient at parts of the road
        :param np.ndarray wind_speeds: (float[N]) speeds of wind in m/s, where > 0 means against the direction of the vehicle
        :param float tick: length of 1 update cycle in seconds
        :param np.ndarray coords: ([float[N,2]) The lat,lon coordinate  of the car at each tick
        :returns: (float[N]) energy expended by the motor at every tick
        :rtype: np.ndarray

        """
        net_force, required_angular_speed_rads = self.calculate_net_force(required_speed_kmh, wind_speeds, gradients)

        cornering_work = self.calculate_cornering_losses(required_speed_kmh, coords, tick)

        motor_output_energies = required_angular_speed_rads * net_force * self.tire_radius * tick + cornering_work
        motor_output_energies = np.clip(motor_output_energies, a_min=0, a_max=None)

        e_m = self.calculate_motor_efficiency(required_angular_speed_rads, motor_output_energies, tick)
        e_mc = self.calculate_motor_controller_efficiency(required_angular_speed_rads, motor_output_energies, tick)

        motor_controller_input_energies = motor_output_energies / (e_m * e_mc)

        # Filter out and replace negative energy consumption as 0
        motor_controller_input_energies = np.where(motor_controller_input_energies > 0,
                                                   motor_controller_input_energies, 0)

        return motor_controller_input_energies

    def calculate_cornering_losses(self, required_speed_kmh, coords, tick):
        """
        Calculate the energy losses due to cornering based on vehicle speed and trajectory.

        :param np.ndarray required_speed_kmh: (float[N]) Required speed array in km/h
        :param np.ndarray coords: (float[N, 2]) Array containing latitude and longitude coordinates of the car at each tick
        :param float tick: Length of one update cycle in seconds
        :returns: (float[N]) Energy loss due to cornering at each tick
        :rtype: np.ndarray
        """
        required_speed_ms = required_speed_kmh / 3.6
        cornering_radii = self.calculate_radii(coords)

        centripetal_lateral_force = self.vehicle_mass * (required_speed_ms ** 2) / cornering_radii
        centripetal_lateral_force = np.clip(centripetal_lateral_force, a_min=0, a_max=10000)

        slip_angles_degrees = self.get_slip_angle_for_tire_force(centripetal_lateral_force)
        slip_angles_radians = np.radians(slip_angles_degrees)
        slip_distances = np.tan(slip_angles_radians) * required_speed_ms * tick

        return slip_distances * centripetal_lateral_force * self.cornering_coefficient

    def calculate_radii(self, coords):
        """
        Calculate the cornering radii for a given set of waypoints.

        :param np.ndarray coords: (float[N, 2]) Array containing latitude and longitude coordinates of the car's path
        :returns: (float[N]) Array of cornering radii at each waypoint
        :rtype: np.ndarray
        """

        # pop off last coordinate if first and last coordinate are the same
        repeated_last_coordinate = False
        if np.array_equal(coords[0], coords[len(coords) - 1]):
            coords = coords[:-1]
            repeated_last_coordinate = True

        cornering_radii = np.empty(len(coords))
        for i in range(len(coords)):
            # if the next point or previous point is out of bounds, wrap the index around the array
            i2 = (i - 1) % len(coords)
            i3 = (i + 1) % len(coords)
            current_point = coords[i]
            previous_point = coords[i2]
            next_point = coords[i3]

            x1 = 0
            y1 = 0
            x2, y2 = self.calculate_meter_distance(current_point, previous_point)
            x3, y3 = self.calculate_meter_distance(current_point, next_point)
            cornering_radii[i] = self.radius_of_curvature(x1, y1, x2, y2, x3, y3)

        # If the last coordinate was removed, duplicate the first radius value to the end of the array
        if repeated_last_coordinate:
            cornering_radii = np.append(cornering_radii, cornering_radii[0])

        # ensure that super large radii are bounded by a large number, like 10000
        cornering_radii = np.where(np.isnan(cornering_radii), 10000, cornering_radii)
        cornering_radii = np.where(cornering_radii > 10000, 10000, cornering_radii)

        return cornering_radii

    def generate_slip_angle_lookup(self, min_degrees, max_degrees, num_elements):
        """
        Generate a lookup table of slip angles and corresponding tire forces using Pacejka's Magic Formula.

        https://www.edy.es/dev/docs/pacejka-94-parameters-explained-a-comprehensive-guide/

        :param float min_degrees: Minimum slip angle in degrees
        :param float max_degrees: Maximum slip angle in degrees
        :param int num_elements: Number of discrete elements in the lookup table
        :returns: (float[num_elements], float[num_elements]) Arrays of slip angles (degrees) and corresponding tire forces (Newtons)
        :rtype: tuple[np.ndarray, np.ndarray]
        """

        b = .25  # Stiffness
        c = 2.2  # Shape
        d = 2.75  # Peak
        e = 1.0  # Curvature

        fz = self.vehicle_mass * 9.81  # Newtons

        slip_angles = np.linspace(min_degrees, max_degrees, num_elements)
        tire_forces = fz * d * np.sin(
            c * np.arctan(b * slip_angles - e * (b * slip_angles - np.arctan(b * slip_angles))))

        return slip_angles, tire_forces

    def get_slip_angle_for_tire_force(self, desired_tire_force):
        slip_angles, tire_forces = self.generate_slip_angle_lookup(0, 70, 100000)

        # Use the numpy interpolation function to find slip angle for the given tire force
        estimated_slip_angle = np.interp(desired_tire_force, tire_forces, slip_angles)

        return estimated_slip_angle

    @staticmethod
    def calculate_meter_distance(coord1: NDArray, coord2: NDArray):
        """
        Calculate the x and y distance in meters between two latitude-longitude coordinates.

        :param tuple coord1: (float[2]) The (latitude, longitude) coordinates of the first point
        :param tuple coord2: (float[2]) The (latitude, longitude) coordinates of the second point
        :returns: (float[2]) The x (longitude) and y (latitude) distances in meters
        :rtype: tuple
        """
        lat1, lon1 = coord1
        lat2, lon2 = coord2

        # Base coordinate
        coord_base = (lat1, lon1)
        # Coordinate for latitude difference (keep longitude the same)
        coord_lat = (lat2, lon1)
        # Coordinate for longitude difference (keep latitude the same)
        coord_long = (lat1, lon2)

        # Calculate y distance (latitude difference)
        y_distance = haversine(coord_base, coord_lat, unit=Unit.METERS)
        # Calculate x distance (longitude difference)
        x_distance = haversine(coord_base, coord_long, unit=Unit.METERS)

        if lat2 < lat1:
            y_distance = -y_distance
        if lon2 < lon1:
            x_distance = -x_distance

        return x_distance, y_distance

    @staticmethod
    def radius_of_curvature(x1, y1, x2, y2, x3, y3):
        """
        Uses the circumcircle the radius of curvature of a circle passing through three points.

        :param float x1: X-coordinate of the first point
        :param float y1: Y-coordinate of the first point
        :param float x2: X-coordinate of the second point
        :param float y2: Y-coordinate of the second point
        :param float x3: X-coordinate of the third point
        :param float y3: Y-coordinate of the third point
        :returns: Radius of curvature of the circle passing through the three points
        :rtype: float
        """
        numerator = np.sqrt(
            ((x3 - x2) ** 2 + (y3 - y2) ** 2) *
            ((x1 - x3) ** 2 + (y1 - y3) ** 2) *
            ((x2 - x1) ** 2 + (y2 - y1) ** 2)
        )

        denominator = 2 * abs(
            ((x2 - x1) * (y1 - y3) - (x1 - x3) * (y2 - y1))
        )

        return numerator / denominator



from abc import ABC, abstractmethod
import numpy as np
from numpy.typing import ArrayLike, NDArray


class BaseGIS(ABC):
    @abstractmethod
    def calculate_closest_gis_indices(self, cumulative_distances) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_path_elevations(self) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_gradients(self, gis_indices) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_time_zones(self, gis_indices) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_path(self) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def calculate_current_heading_array(self) -> np.ndarray:
        raise NotImplementedError

    def calculate_driving_speeds(
            self,
            average_lap_speeds: ArrayLike,
            simulation_dt: int,
            driving_allowed: ArrayLike,
            idle_time: int
    ) -> NDArray[float]:
        raise NotImplementedError



# Radius of the Earth (m)
EARTH_RADIUS = 6371009

# Acceleration caused by gravity (m/s^2)
ACCELERATION_G = 9.81

# Density of Air at 15C and 101kPa (kg/m^3)
AIR_DENSITY = 1.225

# Maximum number of waypoints that can be given to generate route data
MAX_WAYPOINTS = 10

# Solar Irradiance (W/m^2)
SOLAR_IRRADIANCE = 1353

# As we currently have a limited number of API calls(60) every minute with the
# current Weather API, we must shrink the dataset significantly. As the
# OpenWeatherAPI models have a resolution of between 2.5 - 70 km, we will
# go for a resolution of 25km. Assuming we travel at 100km/h for 12 hours,
# 1200 kilometres/25 = 48 API calls
# As the Google Maps API has a resolution of around 40m between points,
# for ASC, we must cull at 625:1 (because 25,000m / 40m = 625)
REDUCTION_FACTOR = 625



import numpy as np
import pandas as pd
import pathlib
import matplotlib.pyplot as plt
from physics.models.battery.kalman_filter import FilteredBatteryModel
from physics.models.battery.battery_config import BatteryModelConfig, load_battery_config, KalmanFilterConfig


# This test requires a voltage.csv and current.csv in the same directory to run
def csv_to_timeseries_tuples(csv_file):
    path = pathlib.Path(__file__).parent / csv_file
    df = pd.read_csv(path)
    df['Time'] = pd.to_datetime(df['Time'])
    return np.array(list(zip(df['Time'].dt.to_pydatetime(), df['Value'])))


def plot_kalman_results(measured_Ut, predicted_Ut_array, predicted_Uc_array, SOC_array, window=None):
    fig, ax = plt.subplots()

    if window is None:
        window = slice(0, len(predicted_Ut_array), 1)

    # ax.plot(predicted_Uoc_array[window], label="Predicted OCV", color="tab:cyan")
    ax.plot(predicted_Ut_array[window], label=r"Filtered Predicted $U_t$", color="orange")
    ax.plot(measured_Ut[window], label=r"Measured $U_t$", color="tab:red")
    ax.plot(predicted_Uc_array[window] + 100, label=r"Predicted $U_c$", color="magenta")
    ax.axhline(y=100, linestyle='dotted', color="magenta")
    # ax.set_ylim(75, 140)
    ax.set_xticks([])
    ax.set_ylabel("Voltage")

    ax2 = ax.twinx()
    ax2.plot(SOC_array[window], color="tab:cyan", label="Filtered SOC")
    ax2.set_ylabel("SOC")

    ax.legend(loc='upper right')
    ax2.legend(loc='lower left')

    plt.title("Impact of Kalman filtering on equivalent-circuit battery modeling ")
    plt.show()


def kalman_filter():
    
    voltage_data = csv_to_timeseries_tuples('voltage.csv')
    current_data = csv_to_timeseries_tuples('current.csv')

    model_config: BatteryModelConfig = load_battery_config(pathlib.Path(__file__).parent / 'battery_config.toml')

    filter_config: KalmanFilterConfig = KalmanFilterConfig(
        model_config,
        process_noise_matrix=np.diag([
            1e-10 * 0.1,
            1e-6 * 0.1
        ]),
        state_covariance_matrix=np.diag(
            [1e-2 * 0.5,
             1e-1]
        ),
        measurement_noise_vector=np.eye(1, dtype=float) * 1e0 * 0.5
    )

    ekf = FilteredBatteryModel(filter_config, initial_SOC=1.04, initial_Uc=0.0)

    SOC_array = np.zeros(len(voltage_data))
    Ut_array = np.zeros(len(voltage_data))
    current_array = np.zeros(len(voltage_data))
    predicted_Ut_array = np.zeros(len(voltage_data))
    predicted_Uc_array = np.zeros(len(voltage_data))
    predicted_Uoc_array = np.zeros(len(voltage_data))

    # This dataset has 0.1s period between measurements
    time_difference = 0.1

    for i in range(int(len(voltage_data))):
        # for i in range(20000):
        # Calculate time difference between current and previous measurements

        Ut = float(voltage_data[i][1])
        I = float(current_data[i][1])

        ekf.predict_then_update(Ut, I, time_difference)

        SOC_array[i] = ekf.SOC
        Ut_array[i] = Ut
        current_array[i] = I
        predicted_Ut_array[i] = ekf.Ut
        predicted_Uc_array[i] = ekf.Uc
        predicted_Uoc_array[i] = ekf._U_oc(ekf.SOC)

    # example usage
    plot_kalman_results(voltage_data[:, 1], predicted_Ut_array, predicted_Uc_array, SOC_array)


if __name__ == '__main__':
    kalman_filter()



from abc import ABC


class BaseBattery(ABC):
    def __init__(self, initial_energy, max_current_capacity, max_energy_capacity,
                 max_voltage, min_voltage, voltage, state_of_charge):
        super().__init__()

        # Constants
        self.max_current_capacity = max_current_capacity  # max capacity of battery (Ah)
        self.max_energy_capacity = max_energy_capacity  # max energy inside battery (Wh)

        self.max_voltage = max_voltage  # maximum battery voltage (V)
        self.min_voltage = min_voltage  # battery cut-off voltage (V)

        # Variables
        self.stored_energy = initial_energy  # energy inside battery (Wh)
        self.state_of_charge = state_of_charge  # battery state of charge
        self.voltage = voltage  # terminal voltage of the battery (V)

        if self.state_of_charge > 0:
            self.empty = False  # 1 if battery is empty, 0 if battery is not empty
        else:
            self.empty = True

    def __str__(self):
        return (f"Battery stored energy: {self.stored_energy:.2f}Wh\n"
                f"Battery state of charge: {self.state_of_charge * 100:.1f}%\n"
                f"Battery voltage: {self.voltage:.2f}V\n")



from physics.environment.meteorology.base_meteorology import BaseMeteorology
from physics.environment.gis.gis import calculate_path_distances
import numpy as np
import physics_rs
from typing import Optional


class IrradiantMeteorology(BaseMeteorology):
    """
    IrradiantMeteorology encapsulates meteorological data that includes
    solar irradiance data, but not cloud cover.

    """

    def __init__(self, race, weather_forecasts):
        self._race = race
        self._raw_weather_data = weather_forecasts

        self._time_dt: Optional[np.ndarray] = None
        self._ghi: Optional[np.ndarray] = None
        self._longitude: Optional[np.ndarray] = None
        self._latitude: Optional[np.ndarray] = None

        self.last_updated_time = self._raw_weather_data[0, 0, 0]

        super().__init__()

    def spatially_localize(self, cumulative_distances: np.ndarray, simplify_weather: bool = False) -> None:
        """

        :param np.ndarray cumulative_distances: NumPy Array representing cumulative distances theoretically achievable for a given input speed array
        :param bool simplify_weather: enable to only use a single weather coordinate (for track races without varying weather)

        """
        # If racing a track race, there is no need for distance calculations. We will return only the origin coordinate
        # This characterizes the weather at every point along the FSGP tracks
        # with the weather at a single coordinate on the track, which is great for reducing the API calls and is a
        # reasonable assumption to make for FSGP only.
        if simplify_weather:
            self._weather_indices = np.zeros_like(cumulative_distances, dtype=int)
            return

        # a list of all the coordinates that we have weather data for
        weather_coords = self._raw_weather_data[:, 0, 1:3]

        # distances between all the coordinates that we have weather data for
        weather_path_distances = calculate_path_distances(weather_coords)
        cumulative_weather_path_distances = np.cumsum(weather_path_distances)

        # makes every even-index element negative, this allows the use of np.diff() to calculate the sum of consecutive
        # elements
        cumulative_weather_path_distances[::2] *= -1

        # contains the average distance between two consecutive elements in the cumulative_weather_path_distances array
        average_distances = np.abs(np.diff(cumulative_weather_path_distances) / 2)

        self._weather_indices = physics_rs.closest_weather_indices_loop(cumulative_distances, average_distances)

    def temporally_localize(self, unix_timestamps, start_time, tick) -> None:
        """

        Takes in an array of indices of the weather_forecast array, and an array of timestamps. Uses those to figure out
        what the weather forecast is at each time step being simulated.

        we only have weather at discrete timestamps. The car however can be in any timestamp in
        between. Therefore, we must be able to choose the weather timestamp that is closest to the one that the car is in
        so that we can more accurately determine the weather experienced by the car at that timestamp.

        For example, imagine the car is at some coordinate (x,y) at timestamp 100. Imagine we know the weather forecast
        at (x,y) for five different timestamps: 0, 30, 60, 90, and 120. Which weather forecast should we
        choose? Clearly, we should choose the weather forecast at 90 since it is the closest to 100. That's what the
        below code is accomplishing.

        :param np.ndarray unix_timestamps: (int[N]) unix timestamps of the vehicle's journey
        :param int start_time: time since the start of the race that simulation is beginning
        :param int tick: length of a tick in seconds
        :returns: a SolcastEnvironment object with time_dt, latitude, longitude, wind_speed, wind_direction, and ghi.
        :rtype: SolcastEnvironment
        """
        forecasts_array = physics_rs.weather_in_time(unix_timestamps.astype(np.int64),
                                                     self._weather_indices.astype(np.int64),
                                                     self._raw_weather_data, 0)

        self._time_dt = forecasts_array[:, 0]
        self._latitude = forecasts_array[:, 1]
        self._longitude = forecasts_array[:, 2]
        self._wind_speed = forecasts_array[:, 3]
        self._wind_direction = forecasts_array[:, 4]
        self._solar_irradiance = forecasts_array[:, 5]

    def calculate_solar_irradiances(self, coords, time_zones, local_times, elevations):
        """
        Calculates the Global Horizontal Irradiance from the Sun, relative to a location
        on the Earth, for arrays of coordinates, times, elevations and weathers
        https://www.pveducation.org/pvcdrom/properties-of-sunlight/calculation-of-solar-insolation
        Note: If local_times and time_zones are both unadjusted for Daylight Savings, the
                calculation will end up just the same

        :param np.ndarray coords: (float[N][lat, lng]) array of latitudes and longitudes
        :param np.ndarray time_zones: (int[N]) time zones at different locations in seconds relative to UTC
        :param np.ndarray local_times: (int[N]) unix time that the vehicle will be at each location. (Adjusted for Daylight Savings)
        :param np.ndarray elevations: (float[N]) elevation from sea level in m
        :returns: (float[N]) Global Horizontal Irradiance in W/m2
        :rtype: np.ndarray

        """
        return self.solar_irradiance
